"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/category.py $
$Id: category.py 30607 2008-03-14 15:21:14Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from durus.persistent import ComputedAttribute
from durus.persistent_dict import PersistentDict
from durus.persistent_list import PersistentList
from qp.lib.spec import add_getters_and_setters, Mixin
from qp.lib.spec import instance, sequence, specify, mapping
from qp.lib.spec import spec, require, string, pattern, both


class Category (DulcineaPersistent):
    """
    A category is a node in a partial ordering.
    """
    name_is = spec(
        (None, both(string, pattern('[a-zA-Z0-9_-]*$'))),
        "unique, URL-friendly identifier for this category")
    label_is = spec(
        (string, None),
        "human-readable string that describes this category; probably but "
        "not necessarily unique")
    description_is = spec(
        (string, None),
        "paragraph-length, human-readable prose description of this category")
    children_is = sequence(instance('Category'), PersistentList)
    parents_is = sequence(instance('Category'), PersistentList)
    _v_expansion_is = ComputedAttribute

    def __init__(self, name=None):
        specify(self, name=name, label=None, description=None,
                children=PersistentList(),
                parents=PersistentList(),
                _v_expansion=ComputedAttribute())

    def __str__(self):
        return self.name or "*unnamed*"

    def add_child(self, child):
        require(child, Category)
        if child.is_ancestor_of(self):
            raise ValueError("Attempt to create circular relationship")
        if child not in self.children:
            self.children.append(child)
            child.parents.append(self)
        self._invalidate_expansion()

    def remove_child(self, child):
        require(child, Category)
        self.children.remove(child)
        child.parents.remove(self)
        self._invalidate_expansion()

    def set_children(self, children):
        require(children, list)
        for child in self.children[:]: # remove_child() modifies self.children
            self.remove_child(child)
        for child in children:
            self.add_child(child)

    def set_parents(self, parents):
        require(parents, list)
        for parent in self.parents[:]: # remove_child() modifies self.parents
            parent.remove_child(self)
        for parent in parents or []:
            parent.add_child(self)

    def _invalidate_expansion(self):
        self._v_expansion.invalidate()
        for parent in self.parents:
            parent._invalidate_expansion()

    def expand(self):
        """() -> { Category : bool }

        Return a set containing this category and its (recursive) children.
        """
        def compute():
            expansion = {self: 1}
            for child in self.children:
                if child not in expansion:
                    expansion.update(child.expand())
            return expansion
        return self._v_expansion.get(compute)

    def get_descendants(self):
        """() -> [ Category ]

        Return list containing this category's (recursive) children.
        """
        descendants = []
        for child in self.children:
            if child not in descendants:
                descendants.append(child)
                descendants += [descendant
                                for descendant in child.get_descendants()
                                if descendant not in descendants]
        return descendants

    def get_leaves(self):
        """() -> [ Category ]

        Return list containing this category or this category's
        (recursive) children that have no children.
        """
        if self.children:
            leaves = {}
            for child in self.children:
                for leaf in child.get_leaves():
                    leaves[leaf] = 1
            return list(leaves.keys())
        else:
            return [self]

    def is_descendant_of(self, other_category):
        """(Category) -> bool

        Is the other category the same as this one or one of the
        descendant categories?
        """
        return self in other_category.expand()

    def is_ancestor_of(self, other_category):
        """(Category) -> bool

        Is this category the same as or a proper descendant of
        other_category?
        """
        return other_category in self.expand()

    def get_label(self):
        return self.label or self.name

add_getters_and_setters(Category)


class Categorized (Mixin):

    categories_is = sequence(Category, PersistentList)

    def __init__(self):
        self.categories = PersistentList()

    def add_category(self, category):
        self.set_categories(self.categories + [category])

    def set_categories(self, categories):
        """(categories : [Category])
        """
        specify(self, categories=PersistentList(categories))

    def get_categories(self):
        """() -> [Category]
        """
        return list(self.categories)

    def get_descendants_of(self, ancestor):
        return [category for category in self.categories
                if ancestor.is_ancestor_of(category)]

    def in_category(self, other_category):
        if other_category is None:
            return 0
        require(other_category, Category)
        other_categories = other_category.expand()
        for category in self.categories:
            if category in other_categories:
                return 1
        return 0

    def in_categories(self, categories):
        """Is self in all of the categories?
        """
        require(categories, [Category])
        for category in categories:
            if not self.in_category(category):
                return 0
        return 1


def remove_ancestors(categories):
    require(categories, [Category])
    no_ancestors = categories[:]
    for category in categories or []:
        for descendant in category.get_descendants():
            if descendant in categories and category in no_ancestors:
                no_ancestors.remove(category)
    return no_ancestors


class CategoryDatabase (Mixin):

    categories_is = mapping({Category.name_is:Category}, PersistentDict)

    def __init__(self):
        assert self.__class__ is not CategoryDatabase, (
            "CategoryDatabase should be a mixin in a persistent subclass")
        self.categories = PersistentDict()

    def add_category(self, category):
        require(category, Category)
        if self.categories.has_key(category.get_name()):
            raise ValueError("Category %r already exists." %
                             category.get_name())
        self.categories[category.get_name()] = category

    def rename_category(self, category, new_name):
        """(category:Category, new_name:string)
        """
        old_name = category.get_name()
        if not self.categories.has_key(old_name):
            raise ValueError("Category %r not found." % old_name)
        if self.categories.has_key(new_name):
            raise ValueError("Category %r already exists." % new_name)
        category.set_name(new_name)
        self.add_category(category)
        del self.categories[old_name]

    def get_category(self, name):
        """(name:string) -> Category

        Retrieve the named category, or None.
        """
        return self.categories.get(name)

    def get_categories(self):
        """() -> [Category]"""
        return list(self.categories.values())




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/util.py $
$Id: util.py 31297 2008-11-19 16:56:46Z rmasse $
"""
from datetime import datetime
from distutils.fancy_getopt import wrap_text
from durus.utils import byte_string
from formatter import AbstractFormatter, DumbWriter, AS_IS
from os.path import isdir, dirname, abspath
from qp.lib.spec import unicode_string
from qpy import stringify, xml, xml_quote
import re
import time
import sys

if sys.version < "3":
    from __builtin__ import unichr
    from StringIO import StringIO
    from htmllib import HTMLParser
    from urllib import urlretrieve, urlopen
    from htmlentitydefs import name2codepoint
else:
    from io import StringIO
    from html.parser import HTMLParser as PlainHTMLParser
    from urllib.request import urlretrieve, urlopen
    # Patch bug in formatter module.
    import formatter
    def list_filter(*args):
        return list(filter(*args))
    formatter.filter = list_filter
    from html.entities import name2codepoint
    unichr = chr

    class HTMLParser (PlainHTMLParser):

        def __init__(self, formatter):
            PlainHTMLParser.__init__(self)
            self.formatter = formatter
            self.nofill = 0
            self.in_head = False
            self.list_stack = []

        # Overridable -- finish processing of start+end tag: <tag.../>
        def handle_startendtag(self, tag, attrs):
            self.handle_starttag(tag, attrs)
            self.handle_endtag(tag)

        # Overridable -- handle start tag
        def handle_starttag(self, tag, attrs):
            tag = tag.lower()
            if self.in_head:
                return
            if tag == 'head':
                self.in_head = True
            elif tag == 'p':
                self.formatter.end_paragraph(1)
            elif tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
                self.formatter.end_paragraph(1)
                self.formatter.push_font((tag, 0, 1, 0))
            elif tag == 'pre':
                self.formatter.end_paragraph(1)
                self.formatter.push_font((AS_IS, AS_IS , AS_IS, 1))
                self.nofill = self.nofill + 1
            elif tag == 'br':
                self.formatter.add_line_break()
            elif tag == 'hr':
                self.formatter.add_hor_rule()
            elif tag == 'blockquote':
                self.formatter.end_paragraph(1)
                self.formatter.push_margin('blockquote')
            elif tag == 'address':
                self.formatter.end_paragraph(0)
                self.formatter.push_font((AS_IS, 1, AS_IS, AS_IS))
            elif tag == 'ul':
                self.formatter.end_paragraph(not self.list_stack)
                self.formatter.push_margin('ul')
                self.list_stack.append(['ul', '*', 0])
            elif tag == 'li':
                self.formatter.end_paragraph(0)
                if self.list_stack:
                    [dummy, label, counter] = top = self.list_stack[-1]
                    top[2] = counter = counter+1
                else:
                    label, counter = '*', 0
                    self.formatter.add_label_data(label, counter)
            elif tag == 'ol':
                self.formatter.end_paragraph(not self.list_stack)
                self.formatter.push_margin('ol')
                label = '1.'
                for a, v in attrs:
                    if a == 'type':
                        if len(v) == 1:
                            v = v + '.'
                        label = v
                self.list_stack.append(['ol', label, 0])
            elif tag == 'dl':
                self.formatter.end_paragraph(1)
                self.list_stack.append(['dl', '', 0])
            elif tag == 'dt':
                self.ddpop()
            elif tag == 'dd':
                self.ddpop()
                self.formatter.push_margin('dd')
                self.list_stack.append(['dd', '', 0])
            elif tag in ('i', 'em', 'cite'):
                self.formatter.push_font((AS_IS, 1, AS_IS, AS_IS))
            elif tag in ('b', 'strong'):
                self.formatter.push_font((AS_IS, AS_IS, 1, AS_IS))
            elif tag in ('tt', 'code'):
                self.formatter.push_font((AS_IS, AS_IS, AS_IS, 1))
            #else: print('start', tag)

        def handle_endtag(self, tag):
            tag = tag.lower()
            if self.in_head:
                if tag == 'head':
                    self.in_head = False
                return
            elif tag == 'p':
                self.formatter.end_paragraph(1)
            elif tag == 'pre':
                self.formatter.end_paragraph(1)
                self.formatter.pop_font()
                self.nofill = max(0, self.nofill - 1)
            elif tag in ('h1', 'h2', 'h3', 'h4', 'h5', 'h6'):
                self.formatter.end_paragraph(1)
                self.formatter.pop_font()
            elif tag == 'blockquote':
                self.formatter.end_paragraph(1)
                self.formatter.push_margin('blockquote')
            elif tag == 'address':
                self.formatter.end_paragraph(0)
                self.formatter.pop_font()
            elif tag in ('ul', 'ol'):
                if self.list_stack:
                    del self.list_stack[-1]
                self.formatter.end_paragraph(not self.list_stack)
                self.formatter.pop_margin()
            elif tag == 'dl':
                self.ddpop(1)
                if self.list_stack:
                    del self.list_stack[-1]
            elif tag in ('i', 'em', 'tt', 'strong', 'b', 'cite', 'code'):
                self.formatter.pop_font()
            #else: print('end', tag)

        def ddpop(self, bl=0):
            self.formatter.end_paragraph(bl)
            if self.list_stack:
                if self.list_stack[-1][0] == 'dd':
                    del self.list_stack[-1]
                    self.formatter.pop_margin()

        def handle_charref(self, name):
            self.formatter.add_literal_data('#' + name + ';')

        def handle_entityref(self, name):
            if name in self.entitydefs:
                self.handle_data(self.entitydefs[name])
            else:
                self.handle_data('&' + name +';')

        def handle_data(self, data):
            if not self.in_head:
                if self.nofill:
                     self.formatter.add_literal_data(data)
                else:
                     self.formatter.add_flowing_data(data)

def as_unicode(s):
    if isinstance(s, unicode_string):
        return s
    elif isinstance(s, byte_string):
        return s.decode('utf-8')
    return unicode_string(s, 'utf-8')

join_unicode = as_unicode('').join
urlretrieve, urlopen, StringIO # quiet checker

class UnicodeStringIO (object):
    """A StringIO that forces all strings to Unicode
    """
    def __init__(self):
        self.buflist = []

    def getvalue(self):
        return join_unicode([as_unicode(line) for line in self.buflist])

    def write(self, s):
        self.buflist.append(s)


entitydefs_as_unichr = {}
for (name, codepoint) in name2codepoint.items():
    entitydefs_as_unichr[name] = unichr(codepoint)

class EntityCharHTMLParser (HTMLParser):
    """An HTMLParser that replaces HTML entities with their Unicode characters.
    """
    entitydefs = entitydefs_as_unichr

def html2txt(text):
    """(any) -> str
    """
    sio = UnicodeStringIO()
    parser = EntityCharHTMLParser(AbstractFormatter(DumbWriter(sio)))
    parser.feed(stringify(text))
    return sio.getvalue()


SAFE_TAGS = ['<p>', '</p>', '<b>', '</b>', '<i>', '</i>', '<ul>',
             '</ul>', '<li>', '</li>', '<br>', '<pre>', '</pre>',
             '<strong>', '</strong>']

def _htmlescape_except(text, safe_tags):
    """(text) -> xml

    Quotes 'text' for use in an HTML page, except for the tags
    listed in 'safe_tags' which are considered safe (e.g. <p>, <b>).
    Both upper and lower case versions of the tags will be applied
    """
    if text is None:
        return None
    text = stringify(xml_quote(text))
    for tag in safe_tags:
        tag_re = re.compile(stringify(xml_quote(tag)), re.IGNORECASE)
        text = tag_re.sub(tag, text)
    return xml(text)

# re to automatically hyperlink email addresses and URLs
_link_re = re.compile(r"""\b(
    # email address
      [\w.-]+ # local part
      @
      [\w.-]+\.[\w]{2,4} # domain
    |
    # URL
      (?:https?://|www\.) # must start with http or www
      [\w.-]+\.[\w]{2,4} # domain
      (?::\d+)? # optional port
      (?:/[\w#$%&+,-./:;=?@\[\]^_|~]*)? # optional path
   )""", re.VERBOSE)

def activate_links(text):
    """(text:str|None) -> xml|None
    Returns an xml_quoted version of text, with things that look like email
    addresses and URLs turned into hyperlinks.
    """
    def _link_replace(m):
        text = url = m.group(0)
        extra = ""
        if text.find("@") == -1:
            if text[-1] in ".,":
                # don't include as part of the URL (easier to handle here
                # than in the regex)
                extra = text[-1]
                text = text[:-1]
            if not text.startswith("http"):
                url = "http://" + text
        else:
            if not text.startswith("mailto"):
                url = "mailto:" + text
        return '<a href="%s">%s</a>%s' % (url, text, extra)

    if text is None:
        return None
    return xml(_link_re.sub(_link_replace, stringify(xml_quote(text))))


def format_text(text, safe_tags=SAFE_TAGS):
    """(text : string, safe_tags : [str]) -> xml

    Convert an ASCII string containing text into HTML.
    The resulting HTML has been reformatted in the following ways:
       * <, &, and > are escaped
       * Things that look like email addresses and URLs will be turned into
         hyperlinks.
       * If safe_tags is defined, it is expected to be a list of tags
         considered to be safe (e.g. <p>, <b>, etc).  These tags will
         not be escaped.
    """
    return activate_links(_htmlescape_except(text, safe_tags))

def sanitize_url(url):
    """(url:string) -> string
    Try to ensure a URL is well-formed, by adding http:// if it isn't present.
    """
    if url is None:
        return None
    if '@' in url and not url.startswith('mailto:'):
        # assume it's an e-mail address
        url = "mailto:" + url
    elif url.find(":") == -1:
        # assume http:// is missing
        url = "http://" + url
    return url

_paragraph_re = re.compile('\n\n+')

def split_paragraphs(text):
    return _paragraph_re.split(text)

def wrap_paragraphs(text):
    """(text) -> string
    Wrap a sequence of paragraphs for output as plain text.
    """
    if text is None:
        return ""
    return '\n\n'.join([wrap_paragraph(paragraph)
                        for paragraph in split_paragraphs(text)])

def wrap_paragraph(text):
    line_length = 70
    if isinstance(text, unicode_string):
        lines = [text.__class__(x, 'utf-8')
                 for x in wrap_text(text.encode('utf-8'), line_length)]
    else:
        lines = wrap_text(text, line_length)
    return '\n'.join(lines)

def insert_paragraph_tags(text):
    """(text:string|None) -> xml|None

    Prepare a text field for display as HTML.
    Currently this just HTML quotes the string and then inserts
    <p> tags at blank lines.
    """
    if text is None:
        return None
    return xml(
        '<p>' +
        _paragraph_re.sub('</p><p>', stringify(xml_quote(text))) +
        '</p>')

def datetime_to_int(date_time):
    """(date_time:datetime) -> int
    Returns the number of seconds since the epoch.
    """
    return int(time.mktime(date_time.timetuple()))

def beginning_of_next_month(date_time):
    """(date_time:datetime) -> datetime
    Return a datetime for the exact beginning of the month
    following the given date_time.
    """
    year = date_time.year
    month = date_time.month + 1
    if month == 13:
        year += 1
        month = 1
    return datetime(year=year, month=month, day=1, hour=0, minute=0,
                    second=0, microsecond=0, tzinfo=date_time.tzinfo)

def is_new(persistent_object):
    """(persistent_object : durus.persistent.Persistent) -> boolean
    """
    return persistent_object._p_connection is None

def get_module_directory(module):
    result = abspath(module.__file__)
    if isdir(result):
        return result
    else:
        return dirname(result)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/test/utest_table.py $
$Id: utest_table.py 27828 2006-01-17 21:04:08Z dbinger $
"""
from dulcinea.ui.table import Table
from dulcinea.common import CommonTest

class Test (CommonTest):

    def a(self):
        table = Table()
        table.tbody_id = 'test_id'
        table.render()
        table.column(a=1)
        table.column(b=2)
        table.render()
        for x in range(1000):
            table.row(a=x, b=x)
        table.render()

if __name__ == '__main__':
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_period.py $
$Id: utest_period.py 31138 2008-09-17 02:43:20Z dbinger $
"""
from datetime import timedelta, datetime
from dulcinea.period import Period
from qp.lib.tz import UTC
from sancho.utest import UTest

class TestPeriod (UTest):

    def a(self):
        p = Period()
        t = datetime.now(UTC)
        assert t in p
        assert p
        assert not p.is_bounded()
        q = Period(t)
        assert t not in q
        assert p.intersection(q) == q
        assert p.intersection(p) == p
        assert q.intersection(q) == q
        assert not q.is_bounded()
        a = Period(None, t)
        assert t - timedelta(seconds=1) in a
        assert t + timedelta(seconds=1) not in a
        b = Period(t, None)
        str(b)
        repr(b)
        assert not a.intersection(b), a.intersection(b)
        assert not b.intersection(a)
        t1 = t + timedelta(seconds=1)
        t2 = t1 + timedelta(seconds=1)
        t3 = t2 + timedelta(seconds=1)
        assert Period(t, t2).intersection(Period(t1, t3)) == Period(t1, t2)

if __name__ == "__main__":
    TestPeriod()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/sort.py $
$Id: sort.py 29354 2007-02-09 22:26:41Z dbinger $
"""
from qpy import stringify

def str_sort(seq):
    """(seq) -> []

    Sort 'seq' by the str() of each element.
    """
    return sorted(seq, key=stringify)

def lexical_sort(seq):
    """(seq) -> []

    Sort 'seq' by the stringify().lower() of each element.
    """
    def stringify_lower(x):
        return stringify(x).lower()
    return sorted(seq, key=stringify_lower)

def attr_sort(seq, attr):
    """(seq, attr : string) -> []

    Sort 'seq' by the attribute 'attr' of each element.
    """
    attr = str(attr)
    def get_attr(v):
        return getattr(v, attr)
    return sorted(seq, key=get_attr)

def method_sort(seq, method):
    """(seq, method : string) -> []

    Sort 'seq' by the result of calling method 'method' on each element.
    """
    method = str(method)
    def meth(v):
        return getattr(v, method)()
    return sorted(seq, key=meth)










"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_sort.py $
$Id: utest_sort.py 31117 2008-09-16 14:55:12Z dbinger $
"""

from sancho.utest import UTest
from dulcinea.sort import str_sort, lexical_sort
from dulcinea.sort import attr_sort, method_sort

class Thing:
    counter = 0

    def __init__(self, name):
        self.name = name
        Thing.counter += 1
        self.id = "T-%03d" % Thing.counter

    def __str__(self):
        return self.name

    def __lt__(self, other):
        return self.id < other.id

    def get_id(self):
        return self.id


def gen(*args):
    for arg in args:
        yield arg


class SortTest (UTest):

    def check_instances(self):
        thing1 = Thing("bob")
        thing2 = Thing("Tim")
        thing3 = Thing("123-abc")
        things = [thing2, thing1, thing3]
        assert str_sort(things) == [thing3, thing2, thing1]
        assert lexical_sort(gen(thing1, thing2, thing3)) == [
            thing3, thing1, thing2]
        assert attr_sort(things, 'id') == [thing1, thing2, thing3]
        assert attr_sort(things, 'name') == [thing3, thing2, thing1]
        assert method_sort(gen(thing2, thing1, thing3), 'get_id') == [
            thing1, thing2, thing3]
        def fun_get_id(x):
            return x.get_id()
        assert sorted(gen(thing2, thing1, thing3), key=fun_get_id) == [
            thing1, thing2, thing3]

if __name__ == "__main__":
    SortTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/inquiry.py $
$Id: inquiry.py 31290 2008-11-17 14:59:38Z dbinger $
"""
from dulcinea.attachable import Attachable
from dulcinea.base import DulcineaPersistent
from dulcinea.contact import Contact
from dulcinea.interaction import Interactable
from dulcinea.partition import PartKeyed, Partition
from dulcinea.timestamped import Timestamped, reverse_timestamp_sorted
from dulcinea.user import get_matching_user
from dulcinea.util import html2txt, StringIO
from qp.lib.spec import add_getters_and_setters, specify, string, match
from qp.mail.send import sendmail
from qp.mail.rfc822_mailbox import parseaddr
from qp.pub.common import get_publisher
import email
import re
import sys


inquiry_re = re.compile(r"(i|inq|inquiry)(\s*#?0*|-)(?P<key>\d+)", re.I)

def get_inquiry_db():
    return get_publisher().get_connection().get_root().get('inquiry_db')


class Inquiry(DulcineaPersistent, PartKeyed, Timestamped, Interactable):

    msg_is = string
    text_is = string
    title_is = string
    from_field_is = string
    message_id_is = string
    user_is = (None, Contact)
    contact_is = (None, Contact)
    previous_inquiry_is = None # forward declaration

    def __init__(self, raw_msg):
        assert raw_msg, "Cannot instantiate an Inquiry with an empty string"
        Interactable.__init__(self)
        Timestamped.__init__(self)
        PartKeyed.__init__(self)
        self.set_msg(raw_msg)
        parsed_msg = self.parse()
        content_type = parsed_msg.get_content_type()
        if parsed_msg.is_multipart():
            text = '\n'.join(
                [sub_msg.get_payload() for sub_msg in parsed_msg.walk()
                 if sub_msg.get_content_type() == 'text/plain'])
            if not text:
               text = html2txt('\n'.join(
                   [sub_msg.get_payload() for sub_msg in parsed_msg.walk()
                    if sub_msg.get_content_type() == 'text/html']))
            if not text:
                text = ('Multipart message with no "text/plain" '
                        'or "text/html" parts')
        elif content_type == 'text/html':
            text = html2txt(parsed_msg.get_payload())
        elif content_type == 'text/plain':
            text = parsed_msg.get_payload()
        elif content_type and content_type.startswith('multipart/'):
            text = 'Badly formed multipart message'
        else:
            text = ('Bad Content type "%s", want ' % content_type +
                    '"multipart/alternative", "text/plain" or "text/html"')
        self.set_text(text)
        self.set_from_field(parsed_msg.get('from', ''))
        subject = parsed_msg.get('subject')
        self.set_title(subject or 'No Subject')
        self.set_user(get_matching_user(self.get_from_address()))
        self.set_contact(None)
        self.set_message_id(parsed_msg.get('message-id', ''))
        self.previous_inquiry = None

    def __eq__(self, other):
        if not type(other) is Inquiry:
            return False
        return (self.get_text() == other.get_text() and
                self.get_title() == other.get_title() and
                self.get_from_field() == other.get_from_field())

    def __hash__(self):
        return id(self)

    def __str__(self):
        return ("inquiry-%s: <%s> %s" % (self.get_key(),
                                         self.get_from_address(),
                                         self.get_title()))[:70]

    def set_msg(self, msg):
        """(msg : unicode_string | byte_string)
        """
        if not match(msg, self.msg_is):
            try:
                msg = msg.decode(self._get_charset(self._parse(msg)))
            except (LookupError, UnicodeDecodeError):
                try:
                    msg = msg.decode('utf8')
                except UnicodeDecodeError:
                    msg = msg.decode('latin1')
        specify(self, msg=msg)

    def set_previous_inquiry(self, previous_inquiry):
        seen = set([self, previous_inquiry])
        if previous_inquiry is self:
            raise ValueError('Previous Inquiry cannot be the same Inquiry')
        inquiry = previous_inquiry
        while 1:
            next = inquiry.get_previous_inquiry()
            if next is None:
                break
            if next in seen:
                raise ValueError('Previous Inquiry cycle detected')
            seen.add(next)
            inquiry = next
        self.previous_inquiry = previous_inquiry

    def get_from_address(self):
        return parseaddr(self.get_from_field())[1]

    def get_address_fields(self):
        parsed = self.parse()
        addresses = [parsed.get(field) for field in ['From', 'To']]
        if parsed.is_multipart(): # try to dig addresses out of forwarded msgs
            for msg in parsed.get_payload():
                if msg.get_content_type() == 'text/plain':
                    possible_forwarded_msg = msg.get_payload()
                    lines = possible_forwarded_msg.split('\n> ')
                    for line in lines:
                        if line.startswith('From:'):
                            addresses.append(line[6:])
        return sorted(set(addresses))

    def _get_key_from_email_header(self, email_header):
        """(email_header : string) -> int | None
        """
        real_name, email_address = parseaddr(email_header)
        if '@' in email_address:
            local_part, domain = email_address.split('@')
        else:
            local_part = email_address
        if '-' in local_part:
            try:
                return int(local_part.split('-')[1])
            except ValueError:
                pass
        return None

    def get_key_from_title(self):
        m = inquiry_re.search(self.get_title() or '')
        if m:
            try:
                return int(m.group('key'))
            except ValueError:
                pass
        return None

    def get_original_inquiry(self):
        """() -> Inquiry | None
        """
        inquiry_db = get_inquiry_db()

        # Search for relations specified in the Subject line
        key = self.get_key_from_title()
        if key is not None:
            inquiry = inquiry_db.get(key)
            if inquiry:
                return inquiry

        # Search for relations of the form "inquiry-#@domain"
        for header in ['to', 'cc']:
            parsed_header = self.get(header)
            if parsed_header:
                for address_field in parsed_header.split(','):
                    key = self._get_key_from_email_header(address_field)
                    if key is not None:
                        inquiry = inquiry_db.get(key)
                        if inquiry:
                            return inquiry

        # Search for inquiry with the new inquiry's In-Reply-To message ID
        in_reply_to = self.get('in-reply-to')
        if in_reply_to:
            inquiry = inquiry_db.get_inquiry_with_message_id(in_reply_to)
            if inquiry:
                return inquiry

        # If this is a forwarded message, search for inquiry with
        # message ID matching that of the enclosed forwarded message
        message_ids = set()
        parsed_msg = self.parse()
        if parsed_msg.is_multipart():
            for sub_msg in parsed_msg.walk():
                message_id = sub_msg.get('message-id')
                if message_id:
                    message_ids.add(message_id)
        for message_id in message_ids:
            if message_id != self.get_message_id():
                inquiry = inquiry_db.get_inquiry_with_message_id(message_id)
                if inquiry:
                    return inquiry

        return None

    def record_interaction(self, user, email, text):
        if not self.get_contact():
            self.set_contact(user)
        return Interactable.record_interaction(self, user, email, text)

    def get(self, header, default=None):
        return self.parse().get(header, default)

    def _get_charset(self, msg):
        """(msg : email.Message)
        """
        if msg.get_charset():
            return msg.get_charset()
        if msg.get_charsets():
            return msg.get_charsets()[0] or ''
        return ''

    def get_charset(self):
        return self._get_charset(self.parse())

    def get_title(self):
        return self.title or self.get_key()

    def has_interactions(self):
        return bool(len(self.get_interaction_history()))

    def is_closed(self):
        return self.get_part_key() == 'closed'

    def is_open(self):
        return self.get_part_key() == 'open'

    def _parse(self, raw_msg):
        """(raw_msg : string) -> email.Message
        """
        return email.message_from_file(StringIO(raw_msg))

    def parse(self):
        """() -> email.Message
        """
        return self._parse(self.get_msg())

Inquiry.previous_inquiry_is = (None, Inquiry)
add_getters_and_setters(Inquiry)


class InquiryDatabase (Partition, Attachable):

    def __init__(self):
        Partition.__init__(self, value_spec=Inquiry,
                           part_key_spec=('open', 'closed'))
        Attachable.__init__(self)

    def get_inquiries(self, sort=True):
        """() -> [Inquiry]
        """
        inquiries = self.itervalues()
        if sort:
            inquiries = reverse_timestamp_sorted(inquiries)
        return list(inquiries)

    def get_inquiry_with_message_id(self, message_id):
        """() -> Inquiry | None

        """
        if message_id:
            for inquiry in self.get_inquiries():
                if inquiry.get_message_id() == message_id:
                    return inquiry
        return None

    def add_inquiry_from_string(self, raw_msg):
        """() -> int

        Add a new Inquiry to InquiryDatabase or add an interaction to an
        existing inquiry if the inquiry is a followup to a previous inquiry.
        """
        new_inquiry = Inquiry(raw_msg)
        unixfrom = new_inquiry.parse().get_unixfrom()
        webmaster_address = get_publisher().get_webmaster_address()
        if not unixfrom:
            sendmail("No unixfrom in inquiry: " + new_inquiry.get_title(),
                     new_inquiry.get_msg(),
                     [webmaster_address],
                     from_addr=webmaster_address)
            return -1
        if get_publisher().get_feedback_domain() not in unixfrom:
            sendmail("Non-local inquiry: " + new_inquiry.get_title(),
                     new_inquiry.get_msg(),
                     [webmaster_address],
                     from_addr=webmaster_address)
            return -1
        if (self.get_part('open') and
            new_inquiry == self.get_part('open').get_max_item()[1]):
            return -1
        spam_status = new_inquiry.get('X-Spam-Status')
        spam_tags = new_inquiry.get('X-Spam-Tags')
        if spam_tags or spam_status and 'SPAM' in spam_status:
            sendmail("Spam inquiry: " + new_inquiry.get_title(),
                     new_inquiry.get_msg(),
                     [webmaster_address],
                     from_addr=webmaster_address)
            return -1
        original_inquiry = new_inquiry.get_original_inquiry()
        if original_inquiry:
            text = ''
            user = get_matching_user(new_inquiry.get_from_address())
            if not user:
                text = 'From: ' + new_inquiry.get_from_field() + '\n\n'
            text += new_inquiry.get_text()
            original_inquiry.record_interaction(user, Inquiry.EMAIL, text)
            return original_inquiry.get_key()
        else:
            self.add(new_inquiry, 'open')
            return new_inquiry.get_key()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_address.py $
$Id: utest_address.py 30251 2007-11-12 22:55:00Z dbinger $
"""
from dulcinea.address import Address, ContactAddress
from qp.lib.spec import get_spec_problems
from sancho.utest import UTest

class AddressTest (UTest):

    def check_init (self):
        Address()
        a=Address(street1="s1", street2="s2")
        assert a.street1 == 's1'
        assert a.street2 == 's2'
        try:
            Address(bogus="goofy")
            assert 0
        except AttributeError: pass

    def check_complete(self):
        addr = ContactAddress(contact_name="Bogus Contact",
                              contact_phone_number="888 555 1212",
                              street1='1320 St.',
                              city='Arlington',
                              state='VA',
                              country_code='US',
                              zip='22201')
        assert get_spec_problems(addr) == []
        assert addr.has_valid_phone_number()
        assert list(addr._check_complete()) == list([])
        addr.contact_phone_number = ''
        assert addr._check_complete() == [('contact_phone_number',
                                           'missing')]
        addr.contact_phone_number = "888 555 1212"
        # Test the checks on the state field
        addr.address.state = ''
        assert addr._check_complete() == [('state', 'missing')]
        addr.address.state = 'VAR'
        assert addr._check_complete() == [('state', 'bad')]
        addr.address.state = 'VB'
        assert addr._check_complete() == [('state', 'bad')]
        addr.address.state = 'VA'
        addr.address.zip = '29999-'
        assert addr._check_complete() == [('zip', 'bad')]
        addr.address.state = 'XX'
        addr.address.zip = None
        assert addr._check_complete() == [('state', 'bad'), ('zip', 'missing')]
        # Will a Canadian province work?
        addr.address.country_code = 'CA'
        addr.address.state = 'QC'
        assert list(addr._check_complete()) == list([])
        addr2 = addr.copy()
        assert addr == addr2
        addr.address.state = 'MA'
        assert addr == addr2 # because the copy is shallow.
        addr.address = addr.address.copy()
        addr.address.state = 'VA'
        assert addr != addr2
        assert addr != 2

    def check_valid (self):
        addr = ContactAddress()
        assert len(addr._check_complete()) > 0
        assert not addr.is_valid()
        addr.address.street1 = "123 Main St."
        assert len(addr._check_complete()) > 0
        assert not addr.is_valid()
        addr.address.city = "Paris"
        addr.address.country_code = "FR"
        assert addr._check_complete() == [('contact_name', 'missing'),
                                          ('contact_phone_number', 'missing')]
        assert not addr.is_valid()
        addr.contact_name = 'foo'
        addr.contact_phone_number = '1112223333'
        assert addr.is_valid()
        assert get_spec_problems(addr) == []

    def check_phone (self):
        for number in ('514-482-7346',
                       '1-514-482-7346',
                       '+1-514-482-7346',
                       '5144827346',
                       '514 482 7346',
                       '(514) 482-7346',
                       '1 (514) 482-7346',
                       '+1 (514) 482-7346',
                       '703-620-8990 x123',
                       '703-620-8990 x 123',
                       '703-620-8990 ext123',
                       '703-620-8990 ext. 123',
                       '703-620-8990 X123',
                       '703-620-8990 X 123',
                       '703-620-8990 Ext123',
                       '703-620-8990 Ext. 123',
                       '+44 (0)20 8681 9603',
                       '+33 01 47 42 20 34',
                       '+33 03 83 27 56 52'):
            addr = ContactAddress(country_code='US',
                                  contact_phone_number=number)
            assert addr.has_valid_phone_number()
        addr = ContactAddress(country_code='US',
                              contact_phone_number='482 7346')
        assert not addr.has_valid_phone_number()
        addr = ContactAddress(country_code='US',
                              contact_phone_number='482- 7346')


if __name__ == "__main__":
    AddressTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/test/utest_property_list.py $
$Id: utest_property_list.py 28932 2006-10-12 20:29:05Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.property import get_property_db
from dulcinea.property.property import Property
from dulcinea.property.property_list import PropertyList
from dulcinea.property.property_template import MasterTemplate
from dulcinea.property.property_type import PropertyType
from dulcinea.physical_value import PhysicalValue
from sancho.utest import raises

class TestPropertyList (CommonTest):

    def _pre(self):
        CommonTest._pre(self)
        if get_property_db() is None:
            return
        get_type = get_property_db().get_property_type
        self.int_type = get_type(PropertyType.ATOMIC,
                                 PropertyType.INT)
        sec = PhysicalValue(1, 's').unit
        self.pv_type = get_type(PropertyType.ATOMIC,
                                PropertyType.PHYSICAL_VALUE,
                                element_sub_type=sec)
        um =  PhysicalValue(1, 'um').unit
        self.pv_um_type = get_type(PropertyType.ATOMIC,
                                PropertyType.PHYSICAL_VALUE,
                                element_sub_type=um)
        self.str_list_type = get_type(PropertyType.LIST,
                                      PropertyType.STRING)
        self.str_int_table_type = get_type(PropertyType.TABLE,
                                           PropertyType.INT,
                                           key_type=PropertyType.STRING)


    def check_basic (self):
        if get_property_db() is None:
            return
        template = MasterTemplate('downtime', self.pv_type)
        property = template.create_value()
        PropertyList([property])
        PropertyList([template])
        PropertyList([property])
        PropertyList([template])
        raises(TypeError, PropertyList, [property, object()])

    def check_filtering (self):
        if get_property_db() is None:
            return
        template = MasterTemplate("a", self.int_type)
        property = Property(MasterTemplate("b", self.pv_type))
        plist = PropertyList([template, property])
        assert list(plist.get_value_properties()) == [property]
        assert list(plist.get_templates()) == [template]
        assert list(plist.get_visible()) == [template, property]
        assert list(plist + []) == list(plist)

if __name__ == "__main__":
    TestPropertyList()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/misc.py $
$Id: misc.py 29610 2007-03-14 21:12:59Z dbinger $
"""
from durus.persistent import PersistentObject
from durus.persistent_list import PersistentList
from qp.lib.spec import specify, Specified
from qp.lib.spec import string, sequence, add_getters_and_setters, spec
from qp.pub.common import get_connection
from qp.pub.user import User

def get_misc_db():
    return get_connection().get_root().get('misc_db')

class MiscDatabase (PersistentObject, Specified):

    motd_is = spec(
        string,
        "Message-of-the-day")
    recent_users_is = spec(
        sequence(User, PersistentList),
        "A list of most recently registered users.")

    recent_users_limit = 10

    def __init__(self):
        specify(self, motd='', recent_users=PersistentList())

    def note_user_registration(self, user):
        self.recent_users.insert(0, user)
        self.recent_users = self.recent_users[:self.recent_users_limit]

add_getters_and_setters(MiscDatabase)



#!/usr/bin/env python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/bin/expire_session.py $
$Id: expire_session.py 29387 2007-02-12 21:04:47Z dbinger $

Delete sessions whose access time is older than the age passed in as an
argument (in hours).

This script is meant to be run by cron.
"""
import sys
from datetime import timedelta
from qp.lib.site import Site
from qp.pub.common import site_now, get_connection

def main (prog, args):
    usage = "usage: %s [site] [expire time]" % prog

    if len(args) != 2:
        sys.exit(usage)

    sessions = Site(args[0]).get_publisher().get_sessions()
    expiration_hours = float(args[1])
    cutoff = site_now() - timedelta(hours=expiration_hours)
    for key, session in sessions.items():
        if session.get_access_time() < cutoff:
            del sessions[key]
    get_connection().commit()

if __name__ == '__main__':
    main(sys.argv[0], sys.argv[1:])



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/script_util.py $
$Id: script_util.py 30177 2007-10-19 11:30:16Z dbinger $
"""
import sys, os, pdb


class Debugger(pdb.Pdb):
    """
    Calling the constructor of this class puts you in the
    interactive debugger.  This is like pdb.post_mortem, except
    that the traceback is printed and the exit command works.
    """
    def __init__(self):
        pdb.Pdb.__init__(self)
        (exc_type, exc_value, tb) = sys.exc_info()
        up = 0
        if tb is None:
            try:
                raise
            except:
                (exc_type, exc_value, tb) = sys.exc_info()
            down = 1
        self.reset()
        while tb.tb_next is not None:
            tb = tb.tb_next
        self.interaction(tb.tb_frame, tb)

    def interaction(self, frame, traceback):
        self.setup(frame, traceback)
        self.curindex = self.curindex - 1
        self.curframe = self.stack[self.curindex][0]
        self.lineno = None
        self.do_explain()
        self.cmdloop()
        self.forget()

    def do_explain(self, arg=None):
        print("\n")
        for stack_entry in self.stack[self.curindex-3:-1]:
            self.print_stack_entry(stack_entry)
        print("")
        self.do_args(None)
        print("")
        self.do_list("%s,13" % (self.curframe.f_lineno - 12))

    def help_explain(self):
        print("""
Print a slice of the stack, args to this function, and list the list the
section of code being executed.
""")
    def do_exit(self, arg):
        os._exit(1)

    def help_exit(self):
        print("""Terminate this process.""")




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/errors.py $
$Id: errors.py 28932 2006-10-12 20:29:05Z dbinger $
"""

class PropertyError (Exception):
    """Base class for all property-related exceptions."""

class PropertyTypeError (PropertyError):
    """Raised on violation of the property type system"""

class ConstraintError (PropertyError):
    """
    Raised when attempting to set a property value in violation
    of the property's constraint.
    """





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/item.py $
$Id: item.py 30928 2008-07-30 20:15:05Z rmasse $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.common import format_date
from dulcinea.timestamped import Timestamped, reverse_timestamp_sorted
from qp.lib.keep import Keyed, Keep
from qp.lib.spec import add_getters_and_setters, string, boolean, init


class Item(DulcineaPersistent, Keyed, Timestamped):

    title_is = (string, None)
    approved_is = boolean

    def __init__(self):
        assert self.__class__ is not Item, "abstract class"
        Keyed.__init__(self)
        Timestamped.__init__(self)
        init(self, approved=False)

    def absorb(self, other):
        assert self.__class__ is other.__class__
        for key, value in other.__getstate__().items():
            if key != 'key':
                setattr(self, key, value)
                
    def get_search_text(self):
        return ' '.join(self.get_search_text_fields())

    def get_search_text_fields(self):
        return [self.get_title() or '', format_date(self.get_timestamp())]

    def is_approved(self):
        return self.approved

    def can_modify(self, user):
        return user.is_admin()

add_getters_and_setters(Item)


class ItemFolder(DulcineaPersistent, Keep):

    def delete(self, item):
        del self.get_mapping()[item.get_key()]

    def get_items(self):
        """() -> [Item]
        """
        return list(self.get_mapping().values())

    def get_recent_items(self, count=None, include_unapproved=False):
        """(count : int = None, include_unapproved : bool = False) -> [Item]

        Return the 'count' most recent items.  If 'include_unapproved' is True,
        both un-approved and approved items are returned.
        The list returned is in reverse-chronological order by timestamp.
        """
        return reverse_timestamp_sorted(
            [item for item in self.get_mapping().itervalues()
             if item.is_approved() or include_unapproved])[:count]
             
    def get_to_oldest_date(self, oldest_date, **kwargs):
        if oldest_date is not None:
            return [item for item in self.get_recent_items(**kwargs)
                    if oldest_date <= item.get_timestamp()]
        return self.get_recent_items(**kwargs)



"""
$Id: link_triple.py 29610 2007-03-14 21:12:59Z dbinger $
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/link_triple.py $
"""
from durus.persistent import PersistentObject
from qp.lib.spec import add_getters_and_setters
from qp.lib.spec import string, Specified
from qp.pub.common import get_publisher

def get_link_triple_db():
    return get_publisher().get_connection().get_root().get('link_triple_db')


class LinkTripleDatabase (PersistentObject, Specified):

    links_is = [(string, string, (string, None))]

    def __init__(self):
        self.links = []

    def __getitem__(self, index):
        return self.links[index]

    def __setitem__(self, index, link):
        self._p_note_change()
        self.links[index] = link

    def __delitem__(self, index):
        self._p_note_change()
        del self.links[index]

    def get(self, index, default=None):
        if 0 <= index < len(self.links):
            return self.links[index]
        else:
            return default

    def insert_link(self, link, index=None):
        self._p_note_change()
        if index is None:
            index = len(self.links)
        assert 0 <= index <= len(self.links), index
        links_after = self.links[index:]
        self.links = self.links[:index]
        self.links.append(link)
        self.links.extend(links_after)

add_getters_and_setters(LinkTripleDatabase)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/test/utest_browse.py $
$Id: utest_browse.py 31153 2008-09-18 15:09:10Z dbinger $
"""
from dulcinea.common import set_test_environment, CommonTest
from dulcinea.ui.browse import ArchiveDirectory, ImageArchiveDirectory
from dulcinea.ui.browse import IndexArchiveDirectory, format_browse_css
from dulcinea.ui.thumbnail import Image
from dulcinea.util import get_module_directory
from durus.utils import as_bytes
import dulcinea.common

def get_test_archive(type):
    test_directory = get_module_directory(dulcinea.ui) + '/test/'
    return test_directory + 'browse_test.%s' % type

def decorate (obj, content, title=None):
    return content

class ArchiveDirectoryTest (CommonTest):

    def _pre(self):
        CommonTest._pre(self)
        def nothing():
            return None
        dulcinea.common.header = nothing
        dulcinea.common.footer = nothing

    def test_misc(self):
        format_browse_css()

    def test_archive_directory_with_zip(self):
        archive_dir = ArchiveDirectory(get_test_archive('zip'),
                                       decorate=decorate)
        set_test_environment()
        assert archive_dir._q_translate('') == '_q_index'
        archive_dir()
        archive_dir._q_traverse(['browse_test', ''])
        archive_dir._q_traverse(['browse_test', 'trek.jpg'])
        archive_dir._q_traverse(['browse_test', 'a', 'trek.jpg'])
        contents = archive_dir._q_traverse(['browse_test', 'b', 'c', 'd.txt'])
        assert contents == as_bytes('Congratulations!\n')
        if Image is not None:
            set_test_environment(QUERY_STRING='40')
            archive_dir._q_traverse(['browse_test', 'a', 'trek.jpg'])

    def test_image_archive_directory_with_zip(self):
        archive_dir = ImageArchiveDirectory(get_test_archive('zip'))
        if Image is not None:
            set_test_environment(QUERY_STRING='40')
            archive_dir._q_traverse(['browse_test', 'a', 'trek.jpg'])

    def test_image_archive_directory_with_tgz(self):
        archive_dir = ImageArchiveDirectory(get_test_archive('tgz'))
        set_test_environment()
        assert archive_dir._q_translate('') == '_q_index'
        archive_dir._q_traverse(['browse_test', ''])
        archive_dir._q_traverse(['browse_test', 'trek.jpg'])
        archive_dir._q_traverse(['browse_test', 'a', 'trek.jpg'])
        fs = archive_dir._q_traverse(['browse_test', 'b', 'c', 'd.txt'])
        contents = [chunk for chunk in fs]
        assert contents == [as_bytes('Congratulations!\n')]
        if Image is not None:
            set_test_environment(QUERY_STRING='40')
            archive_dir._q_traverse(['browse_test', 'a', 'trek.jpg'])

    def test_index_archive_directory_with_tar(self):
        archive_dir = IndexArchiveDirectory(get_test_archive('tar'))
        set_test_environment()
        assert archive_dir._q_translate('') == '_q_index'
        archive_dir._q_traverse(['browse_test', ''])
        archive_dir._q_traverse(['browse_test', 'trek.jpg'])
        archive_dir._q_traverse(['browse_test', 'a', 'trek.jpg'])
        fs = archive_dir._q_traverse(['browse_test', 'b', 'c', 'd.txt'])
        contents = [chunk for chunk in fs]
        assert contents == [as_bytes('Congratulations!\n')]

if __name__ == "__main__":
    ArchiveDirectoryTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_interaction.py $
$Id: utest_interaction.py 27147 2005-08-03 13:59:08Z dbinger $
"""
from dulcinea.interaction import Interactable
from dulcinea.user import DulcineaUser
from sancho.utest import UTest

class TestUser(DulcineaUser, Interactable):

    def __init__(self, id):
        DulcineaUser.__init__(self, id)
        Interactable.__init__(self)

class InteractableTest (UTest):

    def check_interactions(self):
        user = TestUser('guido')
        assert len(user.get_interaction_history()) == 0
        user.record_interaction(user, 'email', 'email sent')
        assert len(user.get_interaction_history()) == 1
        assert user.get_interaction_codes() == user.INTERACTION_CODES

if __name__ == "__main__":
    InteractableTest()




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/cents.py $
$Id: cents.py 31116 2008-09-16 14:54:46Z dbinger $
"""
from qp.lib.spec import require, specify, Specified

class Cents (Specified):

    cents_is = int

    def __init__(self, value=0):
        specify(self, cents=value)

    def get_cents(self):
        return self.cents

    def __str__(self):
        return "$%.02f" % self.get_dollars()

    def get_dollars(self):
        return self.cents / 100.0

    def __add__(self, other):
        require(other, Cents)
        return Cents(self.cents + other.cents)

    def __sub__(self, other):
        require(other, Cents)
        return Cents(self.cents - other.cents)
        
    def __eq__(self, other):
        require(other, Cents)
        return self.cents == other.cents
        
    def __ne__(self, other):
        require(other, Cents)
        return self.cents != other.cents
        
    def __gt__(self, other):
        require(other, Cents)
        return self.cents > other.cents        
        
    def __ge__(self, other):
        require(other, Cents)
        return self.cents >= other.cents                
        
    def __lt__(self, other):
        require(other, Cents)
        return self.cents < other.cents                        
        
    def __le__(self, other):
        require(other, Cents)
        return self.cents <= other.cents                                
                



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_child_process.py $
$Id: utest_child_process.py 27147 2005-08-03 13:59:08Z dbinger $
"""
from dulcinea import child_process
from sancho.utest import UTest

class ChildProcessTest (UTest):

    def check_cat (self):
        p = child_process.execute(('cat',))
        p.write("hello world\n")
        p.stdin.close()
        line = p.readline()
        assert line == 'hello world\n', repr(line)
        assert p.close() == None

    def check_ls (self):
        p = child_process.execute(("ls", "/"))
        assert "bin" in p.read()
        assert p.close() == None

    def check_error (self):
        p = child_process.execute(("false",))
        assert p.close() == 256


if __name__ == "__main__":
    ChildProcessTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/property_list.py $
$Id: property_list.py 30312 2007-11-27 11:26:59Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.property.property import Property
from dulcinea.property.property_template import PropertyTemplate
from qp.lib.spec import either, require, Mixin, add_getters_and_setters
from qp.lib.spec import specify, string_classes


class PropertyList (DulcineaPersistent):

    list_is = [either(Property, PropertyTemplate)]

    def __init__(self, values=None):
        self.list = []
        if values:
            for value in values:
                self.add(value)

    def __copy__(self):
        result = self.__class__()
        result.list = self.list[:]
        return result

    def __eq__(self, other):
        require(other, PropertyList)
        if len(self) != len(other):
            return False
        def as_dict(property_list):
            return dict((p.get_name(), p) for p in property_list)
        return as_dict(self) == as_dict(other)

    def __str__(self):
        return "PropertyList(%s)" % self.list

    def __len__(self):
        return len(self.list)

    def __getitem__(self, index):
        return self.list[index]

    def __add__(self, other):
        return PropertyList(list(self) + list(other))

    def get_names(self):
        """() -> [string]

        Return the list of all names in the list.
        """
        return [property.get_name() for property in self]

    def find(self, name):
        """(name : string) -> any

        Searches the list for a value with name 'name'.  Returns the
        matching value, or None if it couldn't find one.
        """
        for property in self.list:
            if name == property.get_name():
                return property
        require(name, either(*string_classes))
        return None

    def add(self, property):
        """(property : Property|PropertyTemplate)

        Raises ValueError if this list already has a value with the same name.
        """
        require(property, either(Property, PropertyTemplate))
        if self.find(property.get_name()) is not None:
            raise ValueError(
                "list already has a value named %r" % property.get_name())
        self._p_note_change()
        self.list.append(property)

    def __iter__(self):
        for item in self.list:
            yield item

    def get_value_properties(self):
        """Return a new PropertyList containing all properties from the
        current list that have values.
        """
        return PropertyList([property for property in self
            if isinstance(property, Property)])


    def get_templates(self):
        """Return a new PropertyList containing all properties from the
        current list that don't have values.
        """
        return PropertyList([property for property in self
            if isinstance(property, PropertyTemplate)])

    def get_visible(self):
        return PropertyList([property for property in self
            if not property.is_hidden()])


class Propertied (Mixin):

    properties_is = PropertyList

    def __init__(self):
        self.properties = PropertyList()

    def set_properties(self, properties):
        specify(self, properties=PropertyList(properties))

    def find_property(self, name):
        return self.properties.find(name)

    def get_value_properties(self):
        return self.properties.get_value_properties()

    def get_input_templates(self):
        return self.properties.get_templates()

    def find_value_property(self, name):
        for property in self.properties:
            if property.get_name() == name:
                if property.is_input():
                    return None
                else:
                    return property
        return None

    def get_property_value(self, name, default=None):
        p = self.find_value_property(name)
        if p is None or p.value is None:
            return default
        else:
            return p.value

add_getters_and_setters(Propertied)




"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/tolerance.py $
$Id: tolerance.py 30314 2007-11-27 12:15:22Z dbinger $
"""
from dulcinea.base import Copyable
from qp.lib.spec import either, require, add_getters, Specified, number_classes

class Tolerance (Specified, Copyable):
    """
    A tolerance value can be associated with another value
    in order to specify the relative tolerance (in percent) of
    that value.
    """
    lo_is = either(*number_classes)
    hi_is = either(*number_classes)

    def __init__(self, hi=0.0, lo=None):
        assert hi >= 0
        require(hi, float)
        if lo is None:
            lo = hi
        else:
            assert lo >= 0
            require(lo, float)
        self.hi = hi
        if abs(hi - lo) < 1e-9:
            lo = hi
        self.lo = lo

    def __str__(self):
        return self.format()

    def format(self, html=0):
        if html:
            plusminus = '&plusmn;' # a one character version of +/-
        else:
            plusminus = '+/-'
        if self.hi == self.lo:
            return '%s%g%%' % (plusminus, self.hi)
        else:
            return '+%g%%/-%g%%' % (self.hi, self.lo)

    def __hash__(self):
        return hash((self.lo, self.hi))

    def is_symmetric(self):
        """Return true if the high and low magnitudes are the same."""
        return self.hi == self.lo

add_getters(Tolerance)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_partition.py $
$Id: utest_partition.py 30486 2008-01-21 22:28:33Z rmasse $
"""
from dulcinea.partition import PartKeyed, Partition
from dulcinea.base import DulcineaPersistent
from sancho.utest import UTest, raises

class Test (UTest):

    def check_basics(self):
        class B (DulcineaPersistent, PartKeyed):
            junk_is = str
            def __init__(self):
                PartKeyed.__init__(self)
                self.junk = ''
        partition = Partition()
        a = B()
        assert partition.__nonzero__() is False
        raises(AssertionError, partition.get_max_item)
        assert a.get_key() == None
        assert a.get_part_key() == None
        partition.add(a, 'current')
        raises(AssertionError, partition.add, a, 'current')
        b = B()
        b.junk = 1
        raises(TypeError, partition.add, b, 'current')
        raises(TypeError, partition.add, 1, 'current')
        assert a.get_key() == 1
        assert a.get_part_key() == 'current'
        assert list(partition.get_parts()['current'].items()) == [(1, a)]
        assert partition.get(1) == a
        assert partition.__nonzero__() is True
        assert partition.get_max_item() == (1, a)

    def check_part(self):
        class B (DulcineaPersistent, PartKeyed):
            pass
        partition = Partition(part_key_spec=('open', 'closed'))
        a = B()
        b = B()
        c = B()
        partition.add(a, 'open')
        partition.add(b, 'closed')
        raises(TypeError, partition.add, c, 'bad')
        assert list(partition.get_part('open').items()) == [(1, a)]
        assert list(partition.get_part('closed').items()) == [(2, b)]
        assert partition.get_max_item() == (2, b)
        assert partition.get(1) == a
        assert partition.get(2) == b
        assert partition.get(3) == None
        assert partition.get_part('open') is partition.get_parts()['open']

    def check_rm(self):
        class B (DulcineaPersistent, PartKeyed):
            pass
        partition = Partition(part_key_spec=('open', 'closed'))
        a = B()
        b = B()
        partition.add(a, 'open')
        partition.add(b, 'closed')
        assert partition.get(1) == a
        partition.remove(a)
        assert partition.get(1) == None
        raises(KeyError, partition.remove, a)
        assert partition.get(2) == b
        partition.discard(b)
        assert partition.get(2) == None
        partition.discard(b)
        assert partition.get(2) == None

    def check_move(self):
        class B (DulcineaPersistent, PartKeyed):
            pass
        partition = Partition(part_key_spec=('open', 'closed'))
        a = B()
        b = B()
        c = B()
        partition.add(a, 'open')
        assert partition.get(1) == a
        assert a.get_part_key() == 'open'
        assert a.get_key() in partition.get_parts()['open']
        assert 'closed' not in partition.get_parts()
        partition.move(a, 'closed')
        assert 'closed' in partition.get_parts()
        assert 'open' not in partition.get_parts()
        partition.add(c, 'open')
        assert partition.get(1) == a
        assert a.get_part_key() == 'closed'
        assert c.get_key() in partition.get_parts()['open']
        assert a.get_key() not in partition.get_parts()['open']
        assert a.get_key() in partition.get_parts()['closed']
        raises(TypeError, partition.move, a, 'bad')

    def check_iter(self):
        class B (DulcineaPersistent, PartKeyed):
            pass
        partition = Partition(part_key_spec=('open', 'closed'))
        a = B()
        b = B()
        c = B()
        partition.add(a, 'open')
        partition.add(b, 'open')
        partition.add(c, 'closed')
        assert list(partition.itervalues()) == [c, a, b]
        assert list(partition.iterkeys()) == [3, 1, 2]
        assert list(partition.iteritems()) == [(3, c), (1, a), (2, b)]

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/range_value.py $
$Id: range_value.py 31353 2008-11-25 13:44:05Z dbinger $

Provides the RangeValue class.
"""
from dulcinea.base import Copyable
from dulcinea.numeric import near_cmp
from qp.lib.spec import either, Specified, number_classes

class RangeValue (Specified, Copyable):
    """
    A numeric range.
    """
    lo_is = either(*number_classes)
    hi_is = either(*number_classes)

    def __init__(self, lo=0, hi=0):
        self.lo = lo
        self.hi = hi
        if type(self.lo) is not type(self.hi):
            raise TypeError("inconsistent types: %s and %s" %
                (self.lo, self.hi))
        if type(self.lo) not in number_classes:
            raise TypeError("range endpoint must be a number")
        if self.lo > self.hi:
            raise ValueError("backwards range: %s > %s" % (self.lo, self.hi))

    def __str__ (self):
        return "%s .. %s" % (self.lo, self.hi)

    def __hash__(self):
        return hash((self.lo, self.hi))

    def get_min(self):
        return self.lo

    def get_max(self):
        return self.hi

    def get_tuple(self):
        return (self.lo, self.hi)

    def get_type(self):
        return type(self.lo)

    def format(self):
        if self.lo == self.hi:
            return "%g" % self.lo
        else:
            return "%g .. %g" % (self.lo, self.hi)


    # A lot of arithmetic operations aren't supported since it's not clear what
    # they should mean.  Arithmetic operations involving two range values are
    # not supported at all.

    def __add__(self, other):
        if isinstance(other, number_classes):
            return self.__class__(self.lo + other, self.hi + other)
        else:
            return NotImplemented

    __radd__ = __add__

    def __sub__(self, other):
        if isinstance(other, number_classes):
            return self.__class__(self.lo - other, self.hi - other)
        else:
            return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, number_classes):
            return self.__class__(other - self.hi, other - self.lo)
        else:
            return NotImplemented

    def __mul__(self, other):
        if isinstance(other, number_classes):
            return self.__class__(self.lo * other, self.hi * other)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __div__(self, other):
        if isinstance(other, number_classes):
            return self.__class__(self.lo / other, self.hi / other)
        else:
            return NotImplemented
    __truediv__ = __div__

    def __rdiv__(self, other):
        return NotImplemented
    __rtruediv__ = __div__

    def to_float(self):
        return self.__class__(float(self.lo),
                              float(self.hi))

    def __eq__(self, other):
        if isinstance(other, RangeValue):
            if isinstance(self.lo, float):
                return (near_cmp(self.lo, other.lo) == 0 and
                        near_cmp(self.hi, other.hi) == 0)
            else:
                return self.lo == other.lo and self.hi == other.hi
        else:
            return False

    def __ne__(self, other):
        return not self == other

    def __lt__(self, other):
        """
        Comparison is not well-defined for intervals, but this is
        provided anyway so that sorting does not fail.
        """
        if isinstance(other, RangeValue):
            return self.get_min() < other.get_min()
        else:
            return True # arbitrary
            
    def __le__(self, other):
        return self == other or self < other
            
    def __gt__(self, other):
        """
        Comparison is not well-defined for intervals, but this is
        provided anyway so that sorting does not fail.
        """
        if isinstance(other, RangeValue):
            return self.get_min() > other.get_min()
        else:
            return True # arbitrary

    def __ge__(self, other):
        return self == other or self > other               

    def in_range(self, other):
        """Return true if 'other' is inside the range, ie. >= the low end
        and <= the high end.
        """
        if isinstance (other, RangeValue):
            return (near_cmp(self.lo, other.lo) <= 0 and
                    near_cmp(self.hi, other.hi) >= 0)
        elif isinstance(other, number_classes):
            return (near_cmp(self.lo, other) <= 0 and
                    near_cmp(self.hi, other) >= 0)
        else:
            return False









"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_category.py $
$Id: utest_category.py 29576 2007-03-12 17:48:18Z dbinger $

"""
from sancho.utest import UTest, raises
from dulcinea.category import Category, Categorized, remove_ancestors
from qp.lib.spec import get_spec_problems, Specified

class CategoryTest (UTest):

    def check_basic (self):
        self.category = Category()
        assert str(self.category) == '*unnamed*'
        assert not self.category.name_is("ab%")
        assert not self.category.name_is("[ab]")
        assert not self.category.name_is("a\b")
        assert not self.category.name_is("a b")
        assert not self.category.name_is("a/b")
        assert not self.category.name_is(" a/b")
        assert self.category.name_is("__")
        assert self.category.name_is("34")
        assert self.category.name_is("a8")
        assert self.category.name_is("_a8")
        raises(TypeError, self.category.set_name, " ")
        self.category.set_name("boo")
        assert self.category.get_name() == "boo"
        self.category.set_label("Boo")
        assert self.category.get_label() == "Boo"
        self.category.set_description("casual")
        assert self.category.get_description() == "casual"
        assert get_spec_problems(self.category) == []

    def check_categorized (self):
        class C(Specified, Categorized):
            pass
        self.categorized = C()
        assert get_spec_problems(self.categorized) == []
        self.category1 = Category()
        self.category2 = Category()
        raises(TypeError, self.categorized.add_category, None)
        raises(TypeError, self.categorized.add_category, 42)
        self.categorized.add_category(self.category1)
        assert self.categorized.in_category(self.category1)
        assert not self.categorized.in_category(None)
        assert self.categorized.in_categories([self.category1])
        assert set(self.categorized.get_categories()) == set([self.category1])
        assert set(self.categorized.get_descendants_of(self.category1)) == set(
            [self.category1])
        assert set(
            self.categorized.get_descendants_of(self.category2)) == set([])
        assert not self.categorized.in_categories(
            [self.category1, self.category2])

    def check_ancestry (self):
        self.a = Category("a")
        self.b = Category("b")
        self.c = Category("c")
        self.d = Category("d")
        assert set(self.a.get_children()) == set([])
        assert set(self.b.get_parents()) == set([])
        self.a.add_child(self.b)
        assert set(self.a.get_children()) == set([self.b])
        assert set(self.b.get_parents()) == set([self.a])
        assert set(self.a.get_children()) == set([self.b])
        assert set(self.b.get_parents()) == set([self.a])
        raises(ValueError, self.b.add_child, self.a)
        self.b.add_child(self.c)
        assert set(self.c.get_children()) == set([])
        assert set(self.c.get_parents()) == set([self.b])
        # Now we have a -> b -> c.
        assert set(self.a.get_descendants()) == set([self.b, self.c])
        assert set(self.b.get_descendants()) == set([self.c])
        assert set(self.c.get_descendants()) == set([])
        assert self.c.is_descendant_of(self.a)
        assert self.c.is_descendant_of(self.b)
        assert self.c.is_descendant_of(self.c)
        assert self.c.is_ancestor_of(self.c)
        assert not self.d.is_ancestor_of(self.c)
        self.a.set_children([self.b, self.c])
        # a -> b, a-> c, b-> c, d-> b, d-> c
        assert self.c.is_descendant_of(self.a)
        assert self.c.is_descendant_of(self.b)
        assert self.c.is_descendant_of(self.c)
        raises(ValueError, self.a.set_parents, [self.b, self.c])
        assert self.c.is_descendant_of(self.a)
        assert self.c.is_descendant_of(self.b)
        assert self.c.is_descendant_of(self.c)
        self.a.set_parents([self.d])
        assert self.c.is_descendant_of(self.d)
        self.c.set_parents(list(self.c.get_parents()))
        assert set(list(self.a.expand())) == set([self.a, self.b, self.c])
        assert set(list(self.b.expand())) == set([self.b, self.c])
        assert set(list(self.c.expand())) == set([self.c])
        assert set(self.d.expand()) == set([self.a, self.b, self.c, self.d])
        self.a.remove_child(self.c)
        assert self.a.is_ancestor_of(self.c)
        self.b.remove_child(self.c)
        assert not self.a.is_ancestor_of(self.c)
        assert set(list(self.b.expand())) == set([self.b])
        assert set(list(self.a.expand())) == set([self.a, self.b])
        self.a.get_leaves()
        assert remove_ancestors([self.b, self.a]) == [self.b]
        assert remove_ancestors([]) == []
        assert remove_ancestors([self.a, self.b, self.c]) == [self.b, self.c]
        assert remove_ancestors([self.a, self.c]) == [self.a, self.c]

if __name__ == "__main__":
    CategoryTest()




#!/www/python/bin/python
"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_inquiry.py $
$Id: utest_inquiry.py 30722 2008-04-29 17:37:51Z rmasse $
"""
from dulcinea.common import CommonTest
from dulcinea.contact import Contact
from dulcinea.inquiry import Inquiry, InquiryDatabase, inquiry_re
from dulcinea.inquiry import get_inquiry_db
from dulcinea.user import get_matching_user
from qp.pub.common import get_publisher
from sancho.utest import raises

test_msg0 = '''\
From nobody@%s Thu Dec 2 15:48:08 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject: Quote inquiry: from the get-a-quote form
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form
My wafers are 50\xc5 thick'''

test_msg1 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
X-Spam-Status: OK 0.001 
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject: Quote inquiry: from the get-a-quote form
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form
My wafers are 50\xc5 thick'''

test_msg2 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
From: John Smith <john_smith@default.com>
Subject: Followup message #1
X-Spam-Status: UNSURE 0.45 
To: inquiry-%d@default.com
Content-Type: multipart/alternative;
	boundary="----=_NextPart_000_003A_01C20020.B53EF230"

------=_NextPart_000_003A_01C20020.B53EF230
Content-Type: text/plain;
	charset="iso-8859-1"
Content-Transfer-Encoding: 8bit

This is a test for a followup to inquiry-#@domain

------=_NextPart_000_003A_01C20020.B53EF230--
'''

test_msg3 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
From: Mike <mike@default.com>
Subject: Followup message
To: John Smith <john_smith@default.com>
Cc: Joe Public <joe@public.com>, inquiry-%d@default.com
Content-Type: multipart/alternative;
	boundary="----=_NextPart_000_003A_01C20020.B53EF230"

------=_NextPart_000_003A_01C20020.B53EF230
Content-Type: text/html;
	charset="iso-8859-1"
Content-Transfer-Encoding: 8bit

<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 3.2 Final//EN">
<html><body marginheight=5 topmargin=5>
<style>
body, td, font, p, div, blockquote { font-family: arial, verdana, sans-serif; }
td, font, p, div, blockquote { overflow-x: hidden; }
</style>
<font face="Arial, Verdana, sans-serif" size=2>

<p>This is a test for a followup cc inquiry-#@domain and is also MIME</p>

------=_NextPart_000_003A_01C20020.B53EF230--
'''

test_msg4 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
In-Reply-To: <mailman.5670.1139509993.27774.bar@default.com>
From: Boss <boss@anywhere.com>
Subject: Quote inquiry: from the get-a-quote form
To: Worker <worker@anywhere.com>
cc: inquiry@default.com
Content-Type: text/html

Worker, please handle this, <b>--boss</b>

This test finds original inquiry via Message-ID and In-Reply-To headers
and is also text/html as a top-level content type.
'''

test_msg5 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
From: Mike <mike@default.com>
Subject: Followup message
To: John Smith <john_smith@default.com>
Cc: inquiry@default.com
Content-Type: multipart/alternative;
	boundary="----=_NextPart_000_003A_01C20020.B53EF230"

Test for a badly formed multipart message.
'''

test_msg6 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
From: Mike <mike@default.com>
Subject: Followup message
To: John Smith <john_smith@default.com>
Cc: inquiry@default.com
Content-Type: application/vnd.ms-excel

Message with bad content type
'''

test_msg7 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
From: John Smith <john_smith@default.com>
Subject: Here is an image...
To: inquiry@default.com
Content-Type: multipart/mixed;
	boundary="----=_NextPart_000_003A_01C20020.B53EF230"

------=_NextPart_000_003A_01C20020.B53EF230
Content-Type: application/x-msdownload; name="Attachments00.HQX"
Content-Transfer-Encoding: base64
Content-Disposition: attachment; filename="Attachments00.HQX"
    
Multipart message with no "text/plain" or "text/html" parts

------=_NextPart_000_003A_01C20020.B53EF230--
'''

test_msg8 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 15252 invoked by uid 1075); 14 Feb 2006 21:14:57 -0000
From: "John Smith" <john_smith@default.com>
MIME-Version: 1.0
Content-Type: message/rfc822
Content-Description: forwarded message
Content-Transfer-Encoding: 7bit
Message-ID: <17394.18513.390443.634648@foo.default.com>
Date: Tue, 14 Feb 2006 16:14:57 -0500
To: inquiry@default.com
Subject: forwarded message from John Smith

Content-Type: text/plain; charset=utf-8
Return-Path: <webmaster@default.com>
Delivered-To: foo@bar.default.com
Received: (qmail 19083 invoked from network); 14 Feb 2006 21:12:32 -0000
Received: from foo.default.com (HELO smtp.default.com) (10.27.8.141)
\tby 0 with SMTP; 14 Feb 2006 21:12:32 -0000
Received: (qmail 15216 invoked by alias); 14 Feb 2006 21:12:25 -0000
Delivered-To: default.com-john_smith@default.com
Received: (qmail 15214 invoked from network); 14 Feb 2006 21:12:25 -0000
Received: from foo.default.com (132.151.8.3)
\tby 10.27.8.146 with SMTP; 14 Feb 2006 16:12:25 -0500
Received: (qmail 31077 invoked by alias); 14 Feb 2006 21:12:20 -0000
Delivered-To: default.com-john_smith@default.com
Received: (qmail 31075 invoked from network); 14 Feb 2006 21:12:19 -0000
Received: from smtp.default.com (132.151.8.146)
\tby 132.151.8.3 with SMTP; 14 Feb 2006 16:12:19 -0500
Received: (qmail 15207 invoked from network); 14 Feb 2006 21:12:25 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 14 Feb 2006 16:12:24 -0500
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
Precedence: junk
X-No-Archive: yes
From: John Smith <john_smith@default.com>
To: inquiry@default.com
Cc: John Smith <john_smith@default.com>
Subject: Quote inquiry: test

Name: John Smith
Phone: None
Email: john_smith@default.com

test for associate via forwarded message where outer message
content type is message/rfc822
'''

test_msg9 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 15366 invoked by uid 1075); 14 Feb 2006 21:30:09 -0000
From: "John Smith" <john_smith@default.com>
MIME-Version: 1.0
Content-Type: multipart/mixed; boundary="a/WYMEaMM+"
Content-Transfer-Encoding: 7bit
Message-ID: <17394.19425.261116.680890@foo.default.com>
Date: Tue, 14 Feb 2006 16:30:09 -0500
To: john_smith@default.com
cc: inquiry@default.com
Subject: forwarded message from John Smith


--a/WYMEaMM+
Content-Type: text/plain; charset=us-ascii
Content-Description: message body text
Content-Transfer-Encoding: 7bit


you take this.


--a/WYMEaMM+
Content-Type: message/rfc822
Content-Description: forwarded message
Content-Transfer-Encoding: 7bit

Content-Type: text/plain; charset=utf-8
Return-Path: <webmaster@default.com>
Delivered-To: foo@bar.default.com
Received: (qmail 19083 invoked from network); 14 Feb 2006 21:12:32 -0000
Received: from foo.default.com (HELO smtp.default.com) (10.27.8.141)
\tby 0 with SMTP; 14 Feb 2006 21:12:32 -0000
Received: (qmail 15216 invoked by alias); 14 Feb 2006 21:12:25 -0000
Delivered-To: foo.com-john_smith@default.com
Received: (qmail 15214 invoked from network); 14 Feb 2006 21:12:25 -0000
Received: from foo.default.com (132.151.8.3)
\tby 10.27.8.146 with SMTP; 14 Feb 2006 16:12:25 -0500
Received: (qmail 31077 invoked by alias); 14 Feb 2006 21:12:20 -0000
Delivered-To: default.com-john_smith@default.com
Received: (qmail 31075 invoked from network); 14 Feb 2006 21:12:19 -0000
Received: from smtp.default.com (132.151.8.146)
\tby 132.151.8.3 with SMTP; 14 Feb 2006 16:12:19 -0500
Received: (qmail 15207 invoked from network); 14 Feb 2006 21:12:25 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 14 Feb 2006 16:12:24 -0500
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
Precedence: junk
X-No-Archive: yes
From: John Smith <john_smith@default.com>
To: inquiry@default.com
Subject: Quote inquiry: test

Name: John Smith
Phone: None
Email: john_smith@default.com

test for associate via forwarded message where outer message
content type is multipart/mixed 

--a/WYMEaMM+--
'''
test_msg10 = '''\
From nobody@badexample.org Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject: Quote inquiry: from the get-a-quote form
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test for bad unixfrom'''

test_msg11 = '''\
From nobody@%s Thu Dec 2 15:48:08 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject: I%s: from the get-a-quote form
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form'''

test_msg12 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
X-Spam-Tags: CID
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject: Quote inquiry: from the get-a-quote form
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form
My wafers are 50\xc5 thick'''

test_msg13 = '''\
From nobody@%s Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
X-Spam-Status: SPAM 0.999 
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject: Quote inquiry: from the get-a-quote form
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form
My wafers are 50\xc5 thick'''

test_msg14 = '''\
From nobody@foo.com Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
X-Spam-Status: SPAM 0.999 
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
Subject:
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form'''

test_msg15 = '''\
From nobody@foo.com Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Content-Type: text/plain; charset="us-ascii"
X-Spam-Status: SPAM 0.999 
Message-ID: <mailman.5670.1139509993.27774.bar@default.com>
From: John Smith <john_smith@default.com>
To: inquiry@default.com

Name: John Smith
Phone: None
Email: john_smith@default.com

This is a test from the get a quote form'''

test_msg16 = """\
From nobody@foo.com Thu Dec 2 15:48:02 2004
Received: (qmail 17230 invoked from network); 2 Dec 2004 20:48:00 -0000
Received: from localhost (HELO foo.default.com) (127.0.0.1)
\tby 10.27.8.146 with SMTP; 02 Dec 2004 15:48:00 -0500
Mime-Version: 1.0 (Apple Message framework v752.2)
References: <0.2096394811.get_a_quote@foo.com>
Content-Type: multipart/signed; micalg=sha1; boundary=Apple-Mail-9--126546472; protocol="application/pkcs7-signature"
Message-Id: <EFDB708F-7BF5-458E-A398-397DFFA53886@foo.com>
Cc: The Boss <boss@foo.com>
From: The Boss <boss@foo.com>
Subject: Fwd: Hot Prospect
Date: Wed, 14 Nov 2007 16:41:41 -0500
To: Mr Worker <worker@foo.com>
X-Mailer: Apple Mail (2.752.2)

--Apple-Mail-9--126546472
Content-Transfer-Encoding: 7bit
Content-Type: text/plain;
charset=US-ASCII;
format=flowed

Worker,

Can we do this?

-The Boss


Begin forwarded message:

> From: Mr Cool <cool.dude@foo.com>
> Date: November 13, 2007 6:17:38 AM EST
> To: inquiry@foo.com
> Subject: Im a forwarded message
>
> Name: Mr Cool
> Phone: 555-1212
> Email: cool.dude@foo.com
>
> Im a forwarded message.

--Apple-Mail-9--126546472
Content-Transfer-Encoding: base64
Content-Type: application/pkcs7-signature;
name=smime.p7s
Content-Disposition: attachment;
filename=smime.p7s

BOGUSSIGNATUREccAqCAMIACAQExCzAJBgUrDgMCGgUAMIAGCSqGSIb3DQEHAQAAoIIKITCCBMww
ggQ1oAMCAQICEByunWua9OYvIoqj2nRhbB4wDQYJKoZIhvcNAQEFBQAwXzELMAkGA1UEBhMCVVMx
FzAVBgNVBAoTDlZlcmlTaWduLCBJbmMuMTcwNQYDVQQLEy5DbGFzcyAxIFB1YmxpYyBQcmltYXJ=

--Apple-Mail-9--126546472--"""


class InquiryTest(CommonTest):

    def check_re(self):
        assert inquiry_re.match('I123: Subject')
        assert inquiry_re.match('i 23 Subject')
        assert inquiry_re.match('inquiry-23: another subject')
        assert inquiry_re.match('inq 345 - Re: subject subject')
        assert inquiry_re.match('inq-345: good')
        assert inquiry_re.match('Re: inq-345: bad') is None

    def check_inquiry(self):
        if not get_inquiry_db():
            return
        inquiry = Inquiry('my raw msg')
        assert inquiry.get_part_key() == None
        assert inquiry.get_msg() == 'my raw msg'
        inquiry.set_text('my inquiry text')
        assert inquiry.get_text() == 'my inquiry text'
        inquiry.set_title('Test')
        assert inquiry.get_title() == 'Test'
        inquiry.set_from_field('"John Smith" <john_smith@default.com>')
        assert inquiry.get_from_field() == '"John Smith" <john_smith@default.com>'
        user = Contact('John')
        inquiry_user = inquiry.get_user()
        if inquiry_user:
            assert inquiry_user is get_matching_user(
                inquiry.get_from_address())
        inquiry.set_user(user)
        assert inquiry.get_user() is user
        assert inquiry.get_contact() is None
        inquiry.set_contact(user)
        assert inquiry.get_contact() is user
        assert inquiry.get_charset() is ''
        assert Inquiry(test_msg14).get_title() == 'No Subject'
        inquiry15 = Inquiry(test_msg15)
        assert inquiry15.get_title() == 'No Subject'
        inquiry3 = Inquiry(test_msg3)
        assert inquiry3.get_address_fields() == [
            'John Smith <john_smith@default.com>', 'Mike <mike@default.com>']
        raises(ValueError, inquiry3.set_previous_inquiry, inquiry3)
        inquiry3.set_previous_inquiry(inquiry15)
        assert inquiry3.get_previous_inquiry() is inquiry15
        raises(ValueError, inquiry15.set_previous_inquiry, inquiry3)
        inquiry16 = Inquiry(test_msg16)
        assert inquiry16.get_address_fields() == [
            'Mr Cool <cool.dude@foo.com>',
            'Mr Worker <worker@foo.com>',
            'The Boss <boss@foo.com>']

    def check_eq(self):
        if not get_inquiry_db():
            return
        inquiry = Inquiry(test_msg0)
        inquiry2 = Inquiry(test_msg0)
        assert inquiry == inquiry2
        inquiry2 = Inquiry(test_msg1)
        assert inquiry.get_msg() != inquiry2.get_msg()
        assert inquiry == inquiry2
        inquiry2 = Inquiry(test_msg2)
        assert inquiry.get_msg() != inquiry2.get_msg()
        assert not inquiry == inquiry2

    def check_key_searching(self):
        inquiry = Inquiry('John')
        assert inquiry._get_key_from_email_header('any@default.com') == None
        assert inquiry._get_key_from_email_header('R <mail@default.com>') == None
        assert inquiry._get_key_from_email_header('<inquiry@foo.com>') == None
        assert inquiry._get_key_from_email_header('R <inquiry-1@bar.org>') == 1
        assert inquiry._get_key_from_email_header('inquiry-23@foo.biz>') == 23
        assert inquiry._get_key_from_email_header('R <inquiry-A@b.org>') == None

    def check_inquiry_db(self):
        idb = InquiryDatabase()
        domain_name = get_publisher().get_feedback_domain()
        key = idb.add_inquiry_from_string(test_msg1 % domain_name)
        inquiry = idb.get(key)
        assert inquiry.get_part_key() == 'open'
        assert idb.get_part('open').get(inquiry.get_key()) is inquiry
        assert idb.get_part('closed') is None
        idb.move(inquiry, 'closed')
        assert inquiry.get_part_key() == 'closed'
        assert idb.get_part('closed').get(inquiry.get_key()) is inquiry
        assert idb.get_part('open') is None
        assert inquiry.get_from_field() == 'John Smith <john_smith@default.com>'
        assert inquiry.get_from_address() == 'john_smith@default.com'
        assert inquiry.get_charset() == 'us-ascii'
        assert inquiry is idb.get_inquiries()[-1]
        assert inquiry is idb.get_inquiries(sort=1)[-1]
        assert idb.get_inquiry_with_message_id(
            '<mailman.5670.1139509993.27774.bar@default.com>') is inquiry
        assert idb.get_inquiry_with_message_id('no-such-message-id') is None

    def check_insertion(self):
        if get_inquiry_db() is None:
            return
        idb = get_inquiry_db()
        domain_name = get_publisher().get_feedback_domain()
        assert domain_name
        key = idb.add_inquiry_from_string(test_msg1 % domain_name)
        assert idb.get(key).has_interactions() is False
        key2 = idb.add_inquiry_from_string(test_msg0 % domain_name)
        assert key2 == -1, key2
        key2 = idb.add_inquiry_from_string(test_msg2 % (domain_name, key))
        assert key2 == key, "%d != %d" % (key2, key)
        assert idb.get(key).has_interactions() is True
        key2 = idb.add_inquiry_from_string(test_msg3 % (domain_name, key))
        assert key2 == key, "%d != %d" % (key2, key)
        key2 = idb.add_inquiry_from_string(test_msg4 % domain_name)
        assert key2 == key, "%d != %d" % (key2, key)
        key2 = idb.add_inquiry_from_string(test_msg5 % domain_name)
        assert key2 != key, "%d == %d" % (key2, key)
        assert idb.get(key2).get_text() == 'Badly formed multipart message'
        key2 = idb.add_inquiry_from_string(test_msg6 % domain_name)
        assert idb.get(key2).get_text().startswith('Bad Content type')
        key2 = idb.add_inquiry_from_string(test_msg7 % domain_name)
        assert idb.get(key2).get_text() == ('Multipart message with no '
                                            '"text/plain" or "text/html" parts')
        key2 = idb.add_inquiry_from_string(test_msg8 % domain_name)
        assert key2 == key, "%d != %d" % (key2, key)
        key2 = idb.add_inquiry_from_string(test_msg9 % domain_name)
        assert key2 == key, "%d != %d" % (key2, key)
        key2 = idb.add_inquiry_from_string(test_msg10)
        assert key2 == -1, key2
        key2 = idb.add_inquiry_from_string(test_msg11 % (domain_name, key))
        assert key2 == key, "%d != %d" % (key2, key)
        assert idb.add_inquiry_from_string(test_msg12 % domain_name) == -1
        assert idb.add_inquiry_from_string(test_msg13 % domain_name) == -1

if __name__ == "__main__":
    InquiryTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/country.py $
$Id: country.py 30484 2008-01-21 20:41:06Z dbinger $

Map ISO 3166 two-letter country codes to country names.
"""

_country_list = [
    ('AF', 'Afghanistan'),
    ('AL', 'Albania'),
    ('DZ', 'Algeria'),
    ('AS', 'American Samoa'),
    ('AD', 'Andorra'),
    ('AO', 'Angola'),
    ('AI', 'Anguilla'),
    ('AQ', 'Antarctica'),
    ('AG', 'Antigua and Barbuda'),
    ('AR', 'Argentina'),
    ('AM', 'Armenia'),
    ('AW', 'Aruba'),
    ('AU', 'Australia'),
    ('AT', 'Austria'),
    ('AZ', 'Azerbaijan'),
    ('BS', 'Bahamas'),
    ('BH', 'Bahrain'),
    ('BD', 'Bangladesh'),
    ('BB', 'Barbados'),
    ('BY', 'Belarus'),
    ('BE', 'Belgium'),
    ('BZ', 'Belize'),
    ('BJ', 'Benin'),
    ('BM', 'Bermuda'),
    ('BT', 'Bhutan'),
    ('BO', 'Bolivia'),
    ('BA', 'Bosnia and Herzegovina'),
    ('BW', 'Botswana'),
    ('BV', 'Bouvet Island'),
    ('BR', 'Brazil'),
    ('IO', 'British Indian Ocean Territory'),
    ('BN', 'Brunei Darussalam'),
    ('BG', 'Bulgaria'),
    ('BF', 'Burkina Faso'),
    ('BI', 'Burundi'),
    ('KH', 'Cambodia'),
    ('CM', 'Cameroon'),
    ('CA', 'Canada'),
    ('CV', 'Cape Verde'),
    ('KY', 'Cayman Islands'),
    ('CF', 'Central African Republic'),
    ('TD', 'Chad'),
    ('CL', 'Chile'),
    ('CN', 'China'),
    ('CX', 'Christmas Island'),
    ('CC', 'Cocos (Keeling) Islands'),
    ('CO', 'Colombia'),
    ('KM', 'Comoros'),
    ('CG', 'Congo'),
    ('CK', 'Cook Islands'),
    ('CR', 'Costa Rica'),
    ('CI', "Cote d'Ivoire"),
    ('HR', 'Croatia'),
    ('CU', 'Cuba'),
    ('CY', 'Cyprus'),
    ('CZ', 'Czech Republic'),
    ('CD', 'Democratic Republic of the Congo'),
    ('DK', 'Denmark'),
    ('DJ', 'Djibouti'),
    ('DM', 'Dominica'),
    ('DO', 'Dominican Republic'),
    ('TP', 'East Timor'),
    ('EC', 'Ecuador'),
    ('EG', 'Egypt'),
    ('SV', 'El Salvador'),
    ('GQ', 'Equatorial Guinea'),
    ('ER', 'Eritrea'),
    ('EE', 'Estonia'),
    ('ET', 'Ethiopia'),
    ('FK', 'Falkland Islands'),
    ('FO', 'Faroe Islands'),
    ('FJ', 'Fiji'),
    ('FI', 'Finland'),
    ('FR', 'France'),
    ('GF', 'French Guyana'),
    ('TF', 'French Southern Territories'),
    ('GA', 'Gabon'),
    ('GM', 'Gambia'),
    ('GE', 'Georgia'),
    ('DE', 'Germany'),
    ('GH', 'Ghana'),
    ('GI', 'Gibraltar'),
    ('GR', 'Greece'),
    ('GL', 'Greenland'),
    ('GD', 'Grenada'),
    ('GP', 'Guadeloupe'),
    ('GU', 'Guam'),
    ('GT', 'Guatemala'),
    ('GN', 'Guinea'),
    ('GW', 'Guinea-Bissau'),
    ('GY', 'Guyana'),
    ('HT', 'Haiti'),
    ('HM', 'Heard Island and McDonald Islands'),
    ('VA', 'Holy See (Vatican City State)'),
    ('HN', 'Honduras'),
    ('HK', 'Hong Kong'),
    ('HU', 'Hungary'),
    ('IS', 'Iceland'),
    ('IN', 'India'),
    ('ID', 'Indonesia'),
    ('IR', 'Iran'),
    ('IQ', 'Iraq'),
    ('IE', 'Ireland'),
    ('IL', 'Israel'),
    ('IT', 'Italy'),
    ('JM', 'Jamaica'),
    ('JP', 'Japan'),
    ('JO', 'Jordan'),
    ('KZ', 'Kazakhstan'),
    ('KE', 'Kenya'),
    ('KI', 'Kiribati'),
    ('KW', 'Kuwait'),
    ('KG', 'Kyrgyzstan'),
    ('LA', 'Laos'),
    ('LV', 'Latvia'),
    ('LB', 'Lebanon'),
    ('LS', 'Lesotho'),
    ('LR', 'Liberia'),
    ('LY', 'Libya'),
    ('LI', 'Liechtenstein'),
    ('LT', 'Lithuania'),
    ('LU', 'Luxembourg'),
    ('MO', 'Macau'),
    ('MK', 'Macedonia'),
    ('MG', 'Madagascar'),
    ('MW', 'Malawi'),
    ('MY', 'Malaysia'),
    ('MV', 'Maldives'),
    ('ML', 'Mali'),
    ('MT', 'Malta'),
    ('MH', 'Marshall Islands'),
    ('MQ', 'Martinique'),
    ('MR', 'Mauritania'),
    ('MU', 'Mauritius'),
    ('YT', 'Mayotte'),
    ('MX', 'Mexico'),
    ('FM', 'Micronesia'),
    ('MD', 'Moldava'),
    ('MC', 'Monaco'),
    ('MN', 'Mongolia'),
    ('MS', 'Montserrat'),
    ('MA', 'Morocco'),
    ('MZ', 'Mozambique'),
    ('MM', 'Myanmar'),
    ('NA', 'Namibia'),
    ('NR', 'Nauru'),
    ('NP', 'Nepal'),
    ('NL', 'Netherlands'),
    ('AN', 'Netherlands Antilles'),
    ('NC', 'New Caledonia'),
    ('NZ', 'New Zealand'),
    ('NI', 'Nicaragua'),
    ('NE', 'Niger'),
    ('NG', 'Nigeria'),
    ('NU', 'Niue'),
    ('NF', 'Norfolk Island'),
    ('MP', 'Northern Mariana Islands'),
    ('KP', 'North Korea'),
    ('NO', 'Norway'),
    ('OM', 'Oman'),
    ('PK', 'Pakistan'),
    ('PW', 'Palau'),
    ('PS', 'Palestinian Territory'),
    ('PA', 'Panama'),
    ('PG', 'Papua New Guinea'),
    ('PY', 'Paraguay'),
    ('PE', 'Peru'),
    ('PH', 'Philippines'),
    ('PN', 'Pitcairn Island'),
    ('PL', 'Poland'),
    ('PF', 'Polynesia'),
    ('PT', 'Portugal'),
    ('PR', 'Puerto Rico'),
    ('QA', 'Qatar'),
    ('RE', 'Reunion'),
    ('RO', 'Romania'),
    ('RU', 'Russian Federation'),
    ('RW', 'Rwanda'),
    ('SH', 'Saint Helena'),
    ('KN', 'Saint Kitts and Nevis'),
    ('LC', 'Saint Lucia'),
    ('PM', 'Saint Pierre and Miquelon'),
    ('VC', 'Saint Vincent and the Grenadines'),
    ('WS', 'Samoa'),
    ('SM', 'San Marino'),
    ('ST', 'Sao Tome and Principe'),
    ('SA', 'Saudi Arabia'),
    ('SN', 'Senegal'),
    ('SC', 'Seychelles'),
    ('SL', 'Sierra Leone'),
    ('SG', 'Singapore'),
    ('SK', 'Slovakia'),
    ('SI', 'Slovenia'),
    ('SB', 'Solomon Islands'),
    ('SO', 'Somalia'),
    ('ZA', 'South Africa'),
    ('GS', 'South Georgia and the South Sandwich Islands'),
    ('KR', 'South Korea'),
    ('ES', 'Spain'),
    ('LK', 'Sri Lanka'),
    ('SD', 'Sudan'),
    ('SR', 'Suriname'),
    ('SJ', 'Svalbard and Jan Mayen'),
    ('SZ', 'Swaziland'),
    ('SE', 'Sweden'),
    ('CH', 'Switzerland'),
    ('SY', 'Syria'),
    ('TW', 'Taiwan'),
    ('TJ', 'Tajikistan'),
    ('TZ', 'Tanzania'),
    ('TH', 'Thailand'),
    ('TG', 'Togo'),
    ('TK', 'Tokelau'),
    ('TO', 'Tonga'),
    ('TT', 'Trinidad and Tobago'),
    ('TN', 'Tunisia'),
    ('TR', 'Turkey'),
    ('TM', 'Turkmenistan'),
    ('TC', 'Turks and Caicos Islands'),
    ('TV', 'Tuvalu'),
    ('UG', 'Uganda'),
    ('UA', 'Ukraine'),
    ('AE', 'United Arab Emirates'),
    ('UK', 'United Kingdom'),
    ('US', 'United States'),
    ('UY', 'Uruguay'),
    ('UM', 'US Minor Outlying Islands'),
    ('UZ', 'Uzbekistan'),
    ('VU', 'Vanuatu'),
    ('VE', 'Venezuela'),
    ('VN', 'Vietnam'),
    ('VG', 'Virgin Islands (British)'),
    ('VI', 'Virgin Islands (US)'),
    ('WF', 'Wallis and Futuna Islands'),
    ('EH', 'Western Sahara'),
    ('YE', 'Yemen'),
    ('YU', 'Yugoslavia'),
    ('ZM', 'Zambia'),
    ('ZW', 'Zimbabwe')
    ]

# Maps country codes to country names.
_country_name = dict(_country_list)

def get_country_name(country_code):
    """(country_code : string) -> string"""
    return _country_name[country_code]

def get_country_list():
    """() -> [(country_code:string, country_name:string)]"""
    return _country_list

def get_country_codes():
    return list(_country_name.keys())



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_range.py $
$Id: utest_range.py 31341 2008-11-21 16:46:13Z dbinger $
"""
from dulcinea.physical_value import PhysicalValue
from dulcinea.range_value import RangeValue
from sancho.utest import UTest, raises

class RangeTest (UTest):

    def check_init(self):
        r1 = RangeValue(1, 3)
        assert r1.get_tuple() == (1,3)
        assert r1.get_min() == 1
        assert r1.get_max() == 3
        assert r1.get_type() == int
        r = RangeValue(1.0, 3.0)
        assert r.get_tuple() == (1.0,3.0)
        assert list(map(type, r.get_tuple())) == [float,float]


    def check_init_err(self):
        raises(ValueError, RangeValue, 1)
        raises(TypeError, RangeValue, 1, 2, 3)
        raises(TypeError, RangeValue, 'foo')
        raises(TypeError, RangeValue, (1, 'foo'))
        raises(TypeError, RangeValue, (1, 3.0))
        raises(TypeError, RangeValue, (-6.66, 10))
        # range values CANNOT contain PhysicalValues!
        pv1 = PhysicalValue(1.0)
        pv2 = PhysicalValue(2.0)
        raises(TypeError, RangeValue, pv1, pv2)
        raises(ValueError, RangeValue, 1, 0)
        raises(ValueError, RangeValue, 1.5, -5.0)

    def check_format(self):
        r1 = RangeValue(1, 3)
        r2 = RangeValue(2.0, 5.6)
        r3 = RangeValue(0, 0)

        assert str(r1) == "1 .. 3"
        assert r1.format() == "1 .. 3"
        assert str(r2) == "2.0 .. 5.6"
        assert r2.format() == "2 .. 5.6"
        assert r3.format() == "0"


    def check_convert(self):
        r1 = RangeValue(1, 10)
        r2 = RangeValue(1, 10)
        assert r1 == r2
        assert r1.get_type() == int
        assert r2.get_type() == int

        r3 = r2.to_float()
        assert r2.get_type() == int
        assert r3.get_type() == float

        assert r2 == r3
        assert not r2 is r3


    def check_eq(self):
        irange = RangeValue(1, 4)
        frange = RangeValue(2.5, 3.5)
        assert irange == irange
        assert irange == RangeValue(1, 4)
        assert frange == frange
        assert frange == RangeValue(2.5, 3.5)
        assert irange != frange
        assert frange != irange
        assert irange != 37
        assert irange != 'foo'
        assert frange != 37
        assert frange != 'foo'

    def check_addsub(self):
        r1 = RangeValue(1, 4)
        r2 = RangeValue(2, 3)
        assert (r1 + 2).get_tuple() == (3, 6)
        assert (2 + r1).get_tuple() == (3, 6)
        assert (r1 - 2).get_tuple() == (-1, 2)
        assert (2 - r1).get_tuple() == (-2, 1)
        assert (r1 + 2.0).get_tuple() == (3, 6)
        assert (r1 - 2.0).get_tuple() == (-1, 2)
        try:
            r1 + r2
            assert 0
        except TypeError: pass
        try:
            r1 - r2
            assert 0
        except TypeError: pass

    def check_muldiv(self):
        r1 = RangeValue(1, 4)
        r2 = RangeValue(2, 3)
        # Multiplication and division: range with scalar
        assert (r1 * 2).get_tuple() == (2, 8)
        assert (2 * r1).get_tuple() == (2, 8)
        assert (r1 / 2.0).get_tuple() == (0.5, 2)

        fr1 = RangeValue(*map(float, r1.get_tuple()))
        assert fr1 == r1
        assert (fr1 / 2).get_tuple() == (0.5, 2.0)

        # Multiplication and division of ranges
        try:
            r1 * r2
            assert 0
        except TypeError: pass
        try:
            r1 / r2
            assert 0
        except TypeError: pass

        # Try dividing float range
        r1 = r1.to_float()
        assert (r1 / 2.0).get_tuple() == (0.50, 2.0)


    def check_compare(self):
        r1 = RangeValue(1, 4)
        r1a = RangeValue(1.0, 4.0)
        r2 = RangeValue(-2.5, 3.25)

        # ranges can only be compared for equality -- relations
        # don't make sense because of Python's weak comparison
        # semantics ;-(
        assert r1 == RangeValue(1, 4)
        assert r1 == r1a
        assert r1 != r2

        # test the 'in_range()' method
        assert r1.in_range(2)
        assert r1.in_range(1)
        assert r1.in_range(4)
        assert r1.in_range(1.001)
        assert not r1.in_range(0)
        assert not r1.in_range(4.1)

        assert r2.in_range(-2.5)
        assert r2.in_range(3.25)
        assert not r2.in_range(-2.6)
        assert not r2.in_range(3.26)
        list(sorted([2, r1]))


if __name__ == "__main__":
    RangeTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/test/utest_directory.py $
$Id: utest_directory.py 29354 2007-02-09 22:26:41Z dbinger $
"""
from qp.fill.directory import Directory
from sancho.utest import UTest


class DirectoryTest (UTest):

    def b(self):
        class Dir (Directory):
            def get_exports(self):
                return []
        return Dir()._q_translate('') is None

    def c(self):
        class Dir (Directory):
            def get_exports(self):
                return [('')]
        try:
            Dir()._q_translate('') is None
            assert 0
        except ValueError:
            pass

    def d(self):
        class Dir (Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb', 'title')]
        assert Dir()._q_translate('') == '_q_index'

if __name__ == "__main__":
    DirectoryTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/property.py $
$Id: property.py 30544 2008-03-07 20:43:18Z dbinger $
"""
from dulcinea.base import DulcineaPersistent
from dulcinea.property.property_template import PropertyTemplate
from dulcinea.tolerance import Tolerance
from qp.lib.spec import anything, spec, specify, add_getters_and_setters
from qpy import xml, stringify


class Property (DulcineaPersistent):

    template_is = spec(
        PropertyTemplate,
        "defines attributes of this Property such as name, type and "
        "the constraints on the value")
    value_is = spec(
        anything,
        "the value this Property has (may be None which means no value)")
    tolerance_is = spec(
        (None, Tolerance),
        "the relative tolerance of this Property (in percent).  May be "
        "None.")

    def __init__(self, template, value=None):
        specify(self, template=template, tolerance=None)
        self.set_value(value)

    def __str__(self):
        if self.value is None:
            return "%s (no value)" % self.get_name()
        elif not self.value:            # empty list/string/etc.
            return "%s (empty value)" % self.get_name()
        else:
            return "%s = %s" % (self.get_name(), self.format())

    def __eq__(self, other):
        if not isinstance(other, Property):
            return False
        return (self.value == other.value and
                    self.tolerance == other.tolerance and
                    self.get_master() is other.get_master())

    def __ne__(self, other):
        return not (self == other)

    def is_required(self):
        return self.template.is_required()

    def is_hidden(self):
        return self.template.is_hidden()

    def is_input(self):
        return 0

    def is_property(self):
        return 1

    def set_value(self, value):
        """(any)

        May raise TypeError.
        """
        self.template.check_type(value)
        self.value = value

    def allow_tolerance(self):
        return self.template.get_type().allow_tolerance()

    def check_value(self):
        self.template.check_value(self.value)

    def format(self, html=0, show_tolerance=0):
        """Return the Property's value (and unit, if any) as a useful,
        standalone string.  Return None if the value is undefined."""
        if self.value is None:
            return None

        property_type = self.template.get_type()
        s = property_type.format_value(self.value, html=html)
        if show_tolerance and self.tolerance is not None:
            s += ' ' + xml(self.tolerance.format(html=html))
        return s

    def check_type(self):
        """()

        Ensure that 'self.value' is compatible with the Property type.
        Raise PropertyTypeError if not.
        """
        self.get_type().check_value(self.value)

    def simplify(self):
        """() -> (string|int|float|None)
        Return a simplified version of this Property's value. Physical
        values are converted to use the default unit from the master
        template and are returned as floats. Materials are returned as
        strings (using their names).  If self is an InputTemplate,
        simplify the default value
        """
        return self.get_type().simplify_value(self.value)

    # -- Delegates to our template -------------------------------------

    def is_discrete(self):
        return self.template.is_discrete()

    def is_atomic(self):
        return self.template.is_atomic()

    def is_list(self):
        return self.template.is_list()

    def is_table(self):
        return self.template.is_table()

    def is_array(self):
        return self.template.is_array()

    def is_aggregate(self):
        return self.template.is_aggregate()

    def is_time(self):
        return self.template.is_time()

    def get_master(self):
        return self.template.get_master()

    def get_name(self):
        return self.template.get_name()

    def get_type(self):
        return self.template.get_type()

    def get_title(self):
        return self.template.get_title()

    def get_description(self):
        return self.template.get_description()

    def get_hint(self, html=0):
        if html:
            texttype = xml
        else:
            texttype = stringify
        hint = texttype(self.get_description() or "")
        if not (self.is_discrete() or self.get_constraint() is None):
            if hint:
                hint += ', '
            hint += self.template.explain_constraint(html=html)
        return hint

    def get_constraint(self):
        constraint = self.template.get_constraint()
        if self.get_type().is_material():
            materials = set()
            for material in constraint:
                materials.update(material.get_leaves())
            return list(materials)
        else:
            return constraint

    def get_material_constraint(self):
        assert self.get_type().is_material()
        materials = set()
        for material in self.template.get_constraint():
            materials.add(material)
            materials.update(material.get_descendants())
        return list(materials)

    def get_allowed_units(self):
        return self.template.get_allowed_units()

add_getters_and_setters(Property)


def get_example_properties(property_list):
    """() -> [Property]

    Return a list of example value Property instances mirroring 
    the property_list, except that the elements lacking values (e.g. input) 
    are duplicated with values created using the default value of the Property.
    """
    example_properties = []
    if property_list:
        for property in property_list:
            if property.is_input():
                example_properties.append(property.create_value(
                    value=property.get_default_value()))
            else:
                example_properties.append(property.copy())
    return example_properties



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/test/utest_crumbs.py $
$Id: utest_crumbs.py 30177 2007-10-19 11:30:16Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.ui.crumbs import format_crumbs
from qp.fill.directory import Directory, get_crumb_tree
import dulcinea.ui.crumbs


class CrumbTest (CommonTest):

    def test_get_exports1(self):
        exports = [('', '_q_index', 'crumb', 'title')]
        class Dir(Directory):
            def get_exports(self):
                return exports
        assert Dir().get_exports() == exports

    def test_get_crumb_tree(self):
        class Dir3(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb3', 'title3'),
                        ('c3', 'n3', 'acrumb', 'atitle')]

            def _q_traverse(self, components):
                return get_crumb_tree()
            n3 = 'ok'
        class Dir2(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb2', 'title2'),
                        ('c2', 'n2', 'acrumb', 'atitle')]
            n2 = Dir3()
        class Dir1(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb', 'title'),
                        ('c1', 'n1', 'acrumb', 'atitle')]
            n1 = Dir2()

        directory = Dir1()
        assert directory._q_traverse(['c1', 'c2', ''])  == [
            [('./../../', 'crumb', 'title', 'menu_'),
             ('./../../c1', 'acrumb', 'atitle', 'menu_c1')],
            [('./../', 'crumb2', 'title2', 'menu_'),
             ('./../c2', 'acrumb', 'atitle', 'menu_c2')]]

    def test_get_crumb_tree_no_index_crumb_still_has_menu(self):
        class Dir2(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb3', 'title3'),
                        ('c2', 'n2', 'acrumb', 'atitle')]

            def _q_traverse(self, components):
                return get_crumb_tree()
            n2 = 'ok'
        class Dir(Directory):
            def get_exports(self):
                return [('', 'c', '', 'title'),
                        ('c1', 'n1', 'still see crumb', 'still see title')]
            n1 = Dir2()

        directory = Dir()
        assert directory._q_traverse(['c1', 'c2'])  == [
            [('./../c1', 'still see crumb', 'still see title', 'menu_c1')],
            [('./', 'crumb3', 'title3', 'menu_'),
             ('./c2', 'acrumb', 'atitle', 'menu_c2')]]

    def test_format_crumbs(self):
        class Dir3(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb', 'title'),
                        ('c3', 'n3', 'acrumb', 'atitle')]

            def _q_traverse(self, components):
                return format_crumbs()
            n3 = 'ok'
        class Dir2(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb', 'title'),
                        ('c2', 'n2', 'acrumb', 'atitle')]
            n2 = Dir3()
        class Dir1(Directory):
            def get_exports(self):
                return [('', '_q_index', 'crumb', 'title'),
                        ('c1', 'n1', 'acrumb', 'atitle')]
            n1 = Dir2()

        directory = Dir1()
        crumb_menus = False
        def get_config_value(name):
            return crumb_menus
        dulcinea.ui.crumbs.get_config_value = get_config_value
        crumb_menus = True

if __name__ == "__main__":
    CrumbTest()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/property/__init__.py $
$Id: __init__.py 29368 2007-02-12 16:25:16Z rmasse $
"""
from qp.pub.common import get_publisher

def get_property_db():
    return get_publisher().get_connection().get_root().get('property_db')



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/common.py $
$Id: common.py 30600 2008-03-14 11:04:24Z dbinger $
"""
import os
import os.path
import sys
from durus.utils import BytesIO, as_bytes
from qp.lib.site import Site
from qp.pub.common import get_publisher
try:
    from sancho.utest import UTest
except ImportError:
    UTest = object

def set_test_environment(input_string='', site_name=None, **env):
    env.setdefault('SCRIPT_NAME', '')
    env.setdefault('PATH_INFO', '/')
    env.setdefault('SERVER_NAME', 'example.org')
    from qp.pub.hit import Hit
    session_key = "test"
    stdout = sys.stdout
    if site_name is None:
        site_name = os.environ.get('SITE')
    publisher = Site(site_name).get_publisher()
    cookies='%s="%s"' % (publisher.get_site().get_name(), session_key)
    env['HTTP_COOKIE'] = cookies
    hit = Hit(BytesIO(as_bytes(input_string)), env)
    sys.stdout = stdout
    publisher.set_hit(hit)
    session = publisher.create_session()
    publisher.get_sessions()[session_key] = session
    hit.set_session(session)
    def nothing(*args):
        pass
    publisher.commit = nothing
    publisher.abort = nothing

class CommonTest (UTest):

    site_name = None

    def __init__(self, site_name=None):
        self.site_name = site_name
        UTest.__init__(self)

    def _pre(self):
        set_test_environment(site_name=self.site_name)
        get_publisher().get_connection().abort()

def format_user(user, **kwargs):
    return get_publisher().format_user(user, **kwargs)

def format_date(date):
    return get_publisher().format_date(date)

def format_date_time(date):
    return get_publisher().format_date_time(date)

def get_site_title():
    return get_publisher().get_site_title()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_specified_list.py $
$Id: utest_specified_list.py 27147 2005-08-03 13:59:08Z dbinger $
"""
from dulcinea.specified_list import StringList
from sancho.utest import UTest, raises

class Test (UTest):

    def check_stringlist (self):
        s=StringList()
        raises(TypeError, StringList, [1])
        s=StringList(['a'])
        s.append('b')
        assert list(s) == ['a', 'b']
        s.insert(1, 'c')
        assert list(s) == ['a', 'c', 'b']
        s.insert(0, 'd')
        assert list(s) == ['d', 'a', 'c', 'b']
        try:
            s.insert(0, 3)
            assert 0
        except TypeError: pass
        s.insert(7, 'e')
        assert list(s) == ['d', 'a', 'c', 'b', 'e']
        try:
            s[0] = 7
            assert 0
        except TypeError: pass
        s[0] = 'f'
        assert list(s) == ['f', 'a', 'c', 'b', 'e']
        try:
            s[3:4] = ['a', 4]
            assert 0
        except TypeError: pass
        s[3:4] = ['a', 'b']
        assert list(s) == ['f', 'a', 'c', 'a', 'b', 'e']
        raises(TypeError, s.extend, ['a', 4])
        s.extend(['a', 'b'])
        assert list(s) == ['f', 'a', 'c', 'a', 'b', 'e', 'a', 'b']

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/links.py $
$Id: links.py 31152 2008-09-18 11:27:09Z dbinger $
"""
from dulcinea.attachable import Attachable
from dulcinea.category import Category, Categorized, CategoryDatabase
from dulcinea.item import Item, ItemFolder
from dulcinea.util import wrap_paragraphs, sanitize_url
from qp.lib.spec import boolean, spec, init, add_getters_and_setters
from qp.lib.spec import specify, string
from qp.pub.common import get_publisher, site_now
import random

def get_link_db():
    return get_publisher().get_connection().get_root().get('link_db')

class LinkItem(Item, Attachable, Categorized):

    link_url_is = spec(
        (string, None),
        "Link to relevant target URL")
    text_is = spec(
        (string, None),
        "A few sentences describing the link.")
    email_is = spec(
        (string, None),
        "Submitter's e-mail address")
    first_class_is = spec(
        boolean,
        "Flag to indicate that this link can be shown in a prominent "
        "place like the home page of a site")
    old_key_is = spec(
        (None, string),
        "Key used when Items had str keys")

    def __init__(self):
        Item.__init__(self)
        Attachable.__init__(self)
        Categorized.__init__(self)
        init(self, first_class=False)

    def get_allowed_mime_types(self, user=None):
         """() -> [string]
         """
         if user and user.is_admin():
             return Attachable.get_allowed_mime_types(self)
         else:
             return ["image/gif", "image/jpeg", "image/png",
                     "image/tiff", "image/x-ms-bmp"]

    def get_search_text_fields(self):
        return Item.get_search_text_fields(self) + [
            self.get_title(),
            self.get_link_url(),
            self.get_email() or '',
            self.get_text() or ''] + [
            category.get_name() for category in self.get_categories()]

    def set_link_url(self, url):
        specify(self,
                link_url=sanitize_url(url),
                timestamp=site_now())

    def is_first_class(self):
        return self.first_class

    def get_local_url(self):
        "Return the URL to this Item"
        return "/links/%s/" % self.get_key()

    def as_text(self):
        return "%s\n\n%s\nURL: %s" % (self.get_title(), self.get_text(),
                                      self.get_link_url())
    def as_email(self):
        body = "%s\n" % self.get_title()
        if self.get_link_url():
            body += "(%s)\n" % self.get_link_url()
        body += "\n%s" % wrap_paragraphs(self.get_text())
        return self.title, body

add_getters_and_setters(LinkItem)


class LinkFolder(ItemFolder, CategoryDatabase):
    """
    To add a new leaf Category, create the Category instance, set it's 'name'
    and 'label', choose a parent Category from the LinkFolder.categories
    dictionary and call it's 'add_child' to add the new category, finally call
    LinkFolder.add_category to add the new Category to categories
    """

    def __init__(self):
        ItemFolder.__init__(self)
        CategoryDatabase.__init__(self)
        links_category = Category(name='all')
        links_category.set_label('All Links')
        self.add_category(links_category)

    def links_with_category(self, category):
        return [link for link in self.get_items() if link.in_category(category)]

    def get_all_links(self):
        return self.get_items()

    def get_first_class_links(self):
        return [link for link in self.get_items() if link.is_first_class()]

    def random_first_class_link(self):
        first_class_links = self.get_first_class_links()
        if first_class_links:
            return random.choice(first_class_links)
        return None

    def random_first_class_sample(self, size=3):
        first_class_links = self.get_first_class_links()
        if first_class_links:
            return random.sample(first_class_links, size)
        return []

def get_lower_title(link):
    title = link.get_title()
    if title:
        return title.lower()
    else:
        return ''



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/interaction.py $
$Id: interaction.py 29610 2007-03-14 21:12:59Z dbinger $
"""
from dulcinea.history import History
from qp.lib.spec import Mixin

class Interactable (Mixin):
    """
    Storage for a set of interaction events on an object of interest
    """
    interactions_is = History

    EMAIL = 'email'
    MAIL = 'mail'
    PHONE = 'phone'
    FAX = 'fax'
    OTHER = 'other'

    INTERACTION_CODES = (EMAIL, MAIL, PHONE, FAX, OTHER)

    def __init__(self):
        self.interactions = History()

    def get_interaction_history(self):
        return self.interactions

    def get_interaction_codes(self):
        return self.INTERACTION_CODES

    def record_interaction(self, user, code, comment):
        self.interactions.add_event(user, code, comment)





"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/test/utest_note.py $
$Id: utest_note.py 29832 2007-05-29 19:19:08Z dbinger $
"""
from dulcinea.user import DulcineaUser
from dulcinea.note import Note, Notable
from qp.lib.spec import Specified
from qp.pub.common import site_now
from sancho.utest import UTest

class TestNotable (Specified, Notable):
    pass

class Test (UTest):

    def _pre(self):
        self.note = Note()
        self.notable = TestNotable()
        self.user = DulcineaUser("joe")

    def check_note(self):
        assert Note('bar').get_text() == 'bar'
        assert self.note.is_empty()
        assert self.note.get_timestamp() is None
        time = site_now()
        self.note.set_timestamp(time)
        assert self.note.get_timestamp() is time
        assert self.note.get_user() is None
        self.note.set_user(self.user)
        assert self.note.get_user() is self.user
        assert self.note.get_text() is None
        assert self.note.is_empty()
        self.note.set_text('foo')
        assert self.note.get_text() == 'foo'
        assert not self.note.is_empty()

    def check_notable(self):
        assert self.notable.get_note() is None
        self.notable.set_note(self.note)
        assert self.notable.get_note() is self.note
        assert TestNotable(Note('foo')).get_note().get_text() == 'foo'

if __name__ == "__main__":
    Test()



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/calendar.py $
$Id: calendar.py 30485 2008-01-21 21:06:41Z dbinger $
"""
from datetime import timedelta
from dulcinea.base import DulcineaPersistent
from dulcinea.sort import method_sort
from dulcinea.user import DulcineaUser as User
from qp.lib.keep import Keyed, Keep
from qp.lib.spec import boolean, add_getters_and_setters, init, specify
from qp.lib.spec import string, datetime_with_tz


def weekdays(start, end):
    """(start: datetime, end: datetime) -> int

    Return the number of weekdays between two dates.
    """
    days = (end - start).days
    if days < 0:
        raise ValueError('end date must be after start date')
    weeks, partial = divmod(days, 7)
    if partial > 0:
        sunday_distance = (6 - start.weekday()) % 7
        saturday_distance = (5 - start.weekday()) % 7
        if sunday_distance < partial:
            partial -= 1
        if saturday_distance < partial:
            partial -= 1
    return 5*weeks + partial


class Titled(object):

    title_is = string
    description_is = string

    def __init__(self):
        init(self)

add_getters_and_setters(Titled)


class Reservable(Keyed, Titled):

    active_is = boolean

    valid_permissions = {
        'primary_contact' : 'primary contact',
        'control' : 'able to change any reservations of this item',
        'reserve' : 'able to make a reservation of this item',
        }

    def __init__(self):
        Keyed.__init__(self)
        Titled.__init__(self)
        self.active = True

    def set_active(self, value):
        self.active = value

    def is_active(self):
        return self.active

    def allows_reserve(self, user):
        return user.is_granted('reserve', self)

    def allows_control(self, user):
        return (user.is_granted('control', self) or
                user.is_granted('primary_contact', self))


class Resource (DulcineaPersistent, Reservable):

    def __init__(self):
        Reservable.__init__(self)


class Resources (DulcineaPersistent, Keep):

    def __init__(self):
        Keep.__init__(self, Resource)

    def delete_resource(self, resource):
        del self.get_mapping()[resource.get_key()]


class Reservation (DulcineaPersistent, Keyed, Titled):

    start_time_is = datetime_with_tz
    end_time_is = datetime_with_tz
    reserved_is = Reservable
    reserver_is = User

    def __init__(self):
        Keyed.__init__(self)
        Titled.__init__(self)
        init(self)

    def __str__ (self):
        return "%s: %s" % (self.key, self.start_time)

    def set_start_time (self, time):
        time.replace(second=0, microsecond=0)
        specify(self, start_time=time)

    def set_end_time(self, time):
        assert time >= self.start_time
        time.replace(second=0, microsecond=0)
        specify(self, end_time=time)

    def is_active(self, start, end):
        return (self.get_end_time() > start and
                self.get_start_time() < end)

    def can_modify(self, user):
        """(user: User) -> bool
        """
        if user is self.reserver:
            return True
        if not user:
            return False
        if user.is_admin():
            return True
        if self.reserved.allows_control(user):
            return True
        return False

add_getters_and_setters(Reservation)


def create_reservation_table(reservations, time_increment,
                             start_time, end_time,
                             start_time_of_day, end_time_of_day):
    """(reservations : [Reservation]
        time_increment:timedelta,
        start_time:datetime,
        end_time:datetime,
        start_time_of_day:timedelta,
        end_time_of_day:timedelta) -> {
            (day:datetime, time_of_day:timedelta) :
            (reservation:Reservation, rowspan:int) | None }

    This calculates information useful for displaying a tabular view
    of a set of reservations.
    The time_increment is amount of time for each row of the table.
    The start_time and end_time determine the columns (days) of the table.
    The start_time_of_day and end_time_of_day limit the rows that are shown.

    The function returns a dictionary with an entry for each time block
    that is not vacant.  For the first block of each day for a given
    reservation, the dictionary maps (day, time_of_day) to
    (reservation, rowspan), meaning that this reservation appears first
    on this day at this time of day, and that it extends across rowspan
    time increments.

    The cells of the table that are continuations of a reservation that
    has already appeared above have entries that map (day, time_of_day)
    to None.
    """

    reservations = [reservation for reservation in reservations
                    if (reservation.get_start_time() < end_time and
                        reservation.get_end_time() > start_time)]
    reservations = method_sort(reservations, 'get_start_time')
    tableau = {}
    day_delta = timedelta(days=1)
    def gen_days():
        day = start_time
        while day < end_time:
            yield day
            day = day + day_delta
    for day in gen_days():
        day_reservations = [
            reservation for reservation in reservations
            if reservation.is_active(day + start_time_of_day,
                                     day + end_time_of_day)]
        time_of_day = start_time_of_day
        while time_of_day < end_time_of_day:
            if day_reservations:
                reservation = day_reservations[0]
                period_end = day + time_of_day + time_increment
                if reservation.get_start_time() < period_end:
                    start = (day, time_of_day)
                    reservation_end = reservation.get_end_time()
                    rows = 1
                    next_time = time_of_day + time_increment
                    while (next_time < end_time_of_day and
                           day + next_time < reservation_end):
                        rows += 1
                        time_of_day = next_time
                        next_time = next_time + time_increment
                        tableau[(day, time_of_day)] = None # mark continuation
                    tableau[start] = (reservation, rows)
                    day_reservations = day_reservations[1:]
            time_of_day = time_of_day + time_increment
    return tableau



class Calendar (DulcineaPersistent, Keep):

    def __init__(self):
        Keep.__init__(self, Reservation)

    def get_reservations(self):
        return list(self.get_mapping().values())

    def __iter__(self):
        return self.get_mapping().itervalues()

    def delete_reservation(self, reservation):
        del self.get_mapping()[reservation.get_key()]

    def get_reservations_for(self, resource):
        return [reservation for reservation in self.get_reservations()
                if resource is reservation.get_reserved()]

    def get_table_map(self, resource, time_increment,
                      start_time, end_time,
                      start_time_of_day, end_time_of_day):
        return create_reservation_table(self.get_reservations_for(resource),
                                        time_increment, start_time, end_time,
                                        start_time_of_day, end_time_of_day)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/physical_value.py $
$Id: physical_value.py 31142 2008-09-17 13:41:41Z dbinger $
"""
from dulcinea.base import Copyable
from dulcinea.numeric import near_cmp
from dulcinea.physical_unit import get_standard_units
from dulcinea.physical_unit import PhysicalUnit, UnitError
from dulcinea.range_value import RangeValue
from qp.lib.spec import spec, either, add_getters, string, Specified
from qp.lib.spec import number_classes
from qpy import xml, stringify
import operator
import sys

if sys.version < "3":
    div_operator = operator.div
else:
    div_operator = operator.truediv

def _rsub(a, b):
    return b - a

def _rdiv(a, b):
    return b / a


class PhysicalValue (Specified, Copyable):
    """
    A physical value: a number (or range value) tied to a physical
    unit that automatically performs unit conversion when arithmetic
    operations are performed.
    """

    value_is = spec(
        either(float, RangeValue),
        "the numerical part of the physical value")
    unit_is = spec(
        (None, PhysicalUnit),
        "the unit")

    def __init__(self, value=0, unit=None):
        self._set_value(value)
        if unit is None or isinstance(unit, PhysicalUnit):
            self.unit = unit
        elif string(unit):
            # for convenience, only works if get_standard_units is defined
            self.unit = get_standard_units().get_unit(unit)
        else:
            raise ValueError('invalid unit %r' % unit)

    def __str__(self):
        if self.unit:
            return "%s %s" % (self.value, self.unit)
        else:
            return stringify(self.value)

    def __hash__(self):
        return hash((self.value, self.unit))

    def is_range(self):
        """Test if value is a RangeValue"""
        return isinstance(self.value, RangeValue)

    def format(self, html=0, show_unit=1):
        if self.is_range():
            value = self.value.format()
        else:
            value = "%g" % self.value
        if self.unit and show_unit:
            if html: space = xml("&nbsp;")
            else:    space = " "
            result = value + space + self.unit.get_name(html=html)
        else:
            result = value
        return result

    def get_tuple(self):
        return (self.value, self.unit)

    def get_unit_name(self, html=0):
        if self.unit is None:
            return None
        else:
            return self.unit.get_name(html)

    def get_type(self):
        if hasattr(self.value, '__class__'):
            return self.value.__class__
        else:
            return type(self.value)

    def _set_value(self, value):
        # We permit only float and RangeValue of float for the "value"
        # of a PhysicalValue.  Int and range of int both get coerced,
        # to float or range of float respectively. Anything else is a
        # TypeError.

        if value is None:
            raise ValueError("value can't be None")

        if isinstance(value, int):
            self.value = float(value)

        elif isinstance(value, RangeValue):
            rtype = value.get_type()
            if rtype is int:
                self.value = value.to_float() # makes a copy
            elif rtype is float:
                self.value = value
            else:
                raise TypeError("range values must be int or float ranges")

        elif isinstance(value, float):
            self.value = value

        else:
            raise TypeError("invalid value %r for PhysicalValue: "
                "must be int, float, range of int, or range of float" % value)

    def unit_compatible(self, other):
        """(other : PhysicalValue|PhysicalUnit) -> boolean

        Returns true if the PhysicalValue's unit is compatible with
        'other', which may be a PhysicalValue or a PhysicalUnit.
        """
        if self.unit is None:           # self is dimensionless
            unitless_pv = (isinstance(other, PhysicalValue) and
                           other.unit is None)
            dimless_unit = (isinstance(other, PhysicalUnit) and
                            other.is_dimensionless())
            return (unitless_pv or dimless_unit or other is None)

        if isinstance(other, PhysicalValue):
            return self.unit.is_compatible(other.get_unit())
        elif isinstance(other, PhysicalUnit):
            return self.unit.is_compatible(other)
        else:
            raise ValueError('invalid unit %r' % other)


    def is_dimension(self, name):
        if self.unit is not None:
            return self.unit.is_dimension(name)
        else:
            return 0

    def _additive_op(self, other, primitive_op):

        # for arithmetic ops where the units must match, ie. addition
        # and subtraction

        # case 1: adding (subtracting) two PhysicalValue instances
        if isinstance(other, PhysicalValue):
            # case 1a: both self and other have a unit (if they are
            # incompatible, that'll be caught in PhysicalUnit)
            if (self.unit is not None) and (other.unit is not None):
                if self.unit == other.unit:
                    other_value = other.value
                else:
                    other_value = other.unit.convert_value(
                        other.value, self.unit)
                return self.__class__(primitive_op(self.value, other_value),
                                      self.unit)

            # case 1b: neither self nor other has a unit
            elif (self.unit is None) and (other.unit is None):
                return self.__class__(primitive_op(self.value, other.value),
                                      None)

            # case 1c: one of them has a unit, but the other does not
            else:
                raise UnitError("either both values must have a unit, "
                    "or both must be unitless")

        # case 2: adding (subtracting) a PhysicalValue and a scalar or range
        elif isinstance(other, number_classes) or isinstance(other, RangeValue):
            # case 2a: the PhysicalValue is unitless, so it's OK to
            # add/subtract a non-PhysicalValue
            if self.unit is None:
                return self.__class__(primitive_op(self.value, other), None)

            # case 2b: the PhysicalValue has a unit -- bomb!
            else:
                raise UnitError("either both values must be physical values, "
                   "or the physical value must be unitless")

        else:
            return NotImplemented


    def _mult_op(self, other, primitive_op):

        # case 1: multiplying (dividing) two PhysicalValue instances
        if isinstance(other, PhysicalValue):
            assert primitive_op is not _rdiv, \
                   "rdiv impossible: self and other are both PhysicalValues"

            # case 1a: both values have a unit -- don't do any conversion,
            # it's all handled by arithmetic on the units
            if (self.unit is not None) and (other.unit is not None):
                value = primitive_op(self.value, other.value)
                unit = primitive_op(self.unit, other.unit)

                # If all dimensions are zero, then all units cancelled
                # out.  Convert the value to the "standard" no-unit
                # unit, so there's no more useful information lurking in
                # the unit object and we can get rid of it.  (Eg. if we
                # divide "3 m / 1 ft", at this point we'd have "3 m/ft",
                # which is dimensionless but has the conversion factor
                # 3.28 (number of feet in a metre) hidden in the
                # dimensionless "unit".  That's bad -- "3 m / 1 ft" is
                # really the dimensionless quantity "9.84" (= "3 m /
                # .3048 m"), at least if the SI is your fundamental
                # system.  Hence, we convert from some arbitrary
                # dimensionless unit (could have any conversion factor)
                # to the "standard" dimensionless unit (conversion
                # factor = 1).
                #
                # Note that this is different from how we act if the
                # units *don't* cancel out: if you divide miles/hour,
                # you get miles/hour.  This is definitely the right
                # thing to do; it's the units-cancel-out case that's
                # tricky, and I think that ferreting out the hidden
                # conversion factor is the "least-surprise" way of
                # operating.
                if unit.is_dimensionless():
                    value = unit.convert_value(value)
                    unit = None

                return self.__class__(value, unit)

            # case 1b: self is unitless, but other has a unit -- we either
            # use other's unit or its inverse (if dividing)
            elif (self.unit is None) and (other.unit is not None):
                if primitive_op is operator.mul:
                    unit = other.unit
                elif primitive_op is div_operator:
                    unit = other.unit.inv()
                return self.__class__(primitive_op(self.value, other.value),
                                      unit)

            # case 1c: other is unitless, so self's unit doesn't matter
            # (we're just scaling self's value)
            else:
                return self.__class__(primitive_op(self.value, other.value),
                                      self.unit)

        # case 2: multiplying a PhysicalValue by a number or
        # range -- this is just like 1c, except we use 'other'
        # instead of 'other.value'
        elif isinstance(other, number_classes) or isinstance(other, RangeValue):
            if primitive_op is _rdiv:
                unit = self.unit.inv()
            else:
                unit = self.unit
            return self.__class__(primitive_op(self.value, other), unit)

        else:
            return NotImplemented

    def __add__(self, other):
        return self._additive_op(other, operator.add)

    __radd__ = __add__                  # addition is still commutative!

    def __sub__(self, other):
        return self._additive_op(other, operator.sub)

    def __rsub__(self, other):
        return self._additive_op(other, _rsub)

    def __mul__(self, other):
        return self._mult_op(other, operator.mul)

    __rmul__ = __mul__                  # yep, this one's commutative too

    def __div__(self, other):
        return self._mult_op(other, div_operator)
    __truediv__ = __div__

    def __rdiv__(self, other):
        return self._mult_op(other, _rdiv)
    __rtruediv__ = __div__

    def __eq__(self, other):
        other = self.get_comparable_value(other)
        if isinstance(self.value, RangeValue) or isinstance(other, RangeValue):
            return self.value == other
        else:
            return 0 == near_cmp(self.value, other)

    def __ne__(self, other):
        return not self == other

    def __gt__(self, other):
        other = self.get_required_comparable_value(other)
        if isinstance(self.value, RangeValue) or isinstance(other, RangeValue):
            return self.value > other
        else:
            return 1 == near_cmp(self.value, other)

    def __le__(self, other):
        return not self > other

    def __lt__(self, other):
        other = self.get_required_comparable_value(other)
        if isinstance(self.value, RangeValue) or isinstance(other, RangeValue):
            return self.value < other
        else:
            return -1 == near_cmp(self.value, other)

    def __ge__(self, other):
        return not self < other

    def get_comparable_value(self, other):
        """( any ) -> float | int | long | None

        Return a number that represents the other,
        expressed in the units of self.
        If there isn't any sensible number to return, return None.
        """
        if isinstance(other, PhysicalValue):
            if self.unit is other.unit:
                return other.value
            elif (self.unit is not None and
                  other.unit is not None and
                  self.unit.powers == other.unit.powers):
                return other.unit.convert_value(other.value, self.unit)
        elif self.unit is None and isinstance(other, number_classes):
            return other
        return None

    def get_required_comparable_value(self, other):
        """( any ) -> float | int | long

        Return a number that represents the other,
        expressed in the units of self.
        If there isn't any sensible number to return,
        raise an UnitError exception.
        """
        val = self.get_comparable_value(other)
        if val is None:
            raise UnitError("can't compare %r to %r" % (self, other))
        return val

    def convert(self, target_unit):
        """(target_unit : PhysicalUnit) -> PhysicalValue

        Return a (possibly new) PhysicalValue instance with the current
        instance's value converted to 'target_unit'.
        """
        if self.unit is target_unit:
            return self
        return self.__class__(self.unit.convert_value(self.value, target_unit),
                              target_unit)

    def get_min(self):
        if self.is_range():
            return PhysicalValue(self.value.get_min(), self.unit)
        return self

    def get_max(self):
        if self.is_range():
            return PhysicalValue(self.value.get_max(), self.unit)
        return self

    def in_range(self, other):
        if self.is_range():
            return self.value.in_range(self.get_comparable_value(other))
        return self == other

add_getters(PhysicalValue)



"""
$URL: svn+ssh://svn.mems-exchange.org/repos/trunk/dulcinea/lib/ui/test/utest_util.py $
$Id: utest_util.py 29354 2007-02-09 22:26:41Z dbinger $
"""
from dulcinea.common import CommonTest
from dulcinea.ui.util import csv
from qp.pub.common import get_response

class MiscTest (CommonTest):

    def check_csv (self):
        s = csv(['a','b'], [{'a':1}, {'a':2, 'b':3}])
        assert s == 'a,b\r\n1,\r\n2,3\r\n'
        response = get_response()
        mime_type = response.get_mime_type()
        assert mime_type == 'text/comma-separated-values'

if __name__ == "__main__":
    MiscTest()
